import logging
from typing import Dict, Type

from overhave.entities.settings import OverhaveRedisSettings
from overhave.redis.objects import BaseRedisTask, RedisStream
from overhave.redis.template import RedisTemplate

logger = logging.getLogger(__name__)


class RedisProducer(RedisTemplate):
    """ Class for producing tasks.

    Producer send tasks into Redis stream specified by ```mapping``.
    """

    def __init__(self, settings: OverhaveRedisSettings, mapping: Dict[Type[BaseRedisTask], RedisStream]):
        super().__init__(settings)
        self._streams = {task: self._database.Stream(stream.value) for task, stream in mapping.items()}

    def add(self, task: BaseRedisTask) -> None:
        stream = self._streams[type(task)]
        logger.info("Added BaseRedisTask %s", task)
        stream.add(task.message)



from typing import Optional

from overhave.base_settings import OverhaveLoggingSettings
from overhave.entities import (
    OverhaveEmulationSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveScenarioCompilerSettings,
    ProcessorSettings,
)
from overhave.entities.authorization import AuthorizationStrategy
from overhave.entities.authorization.settings import (
    OverhaveAdminSettings,
    OverhaveAuthorizationSettings,
    OverhaveLdapClientSettings,
)
from overhave.stash.settings import OverhaveStashClientSettings, OverhaveStashProjectSettings
from overhave.testing import OverhaveDescriptionManagerSettings, OverhaveProjectSettings, OverhaveTestSettings


class OverhaveContext:
    """ Application Context, based on BaseSettings.

    This context defines how Overhave application will work.
    """

    def __init__(
        self,
        admin_settings: Optional[OverhaveAdminSettings] = None,
        auth_settings: Optional[OverhaveAuthorizationSettings] = None,
        compilation_settings: Optional[OverhaveScenarioCompilerSettings] = None,
        emulation_settings: Optional[OverhaveEmulationSettings] = None,
        file_settings: Optional[OverhaveFileSettings] = None,
        language_settings: Optional[OverhaveLanguageSettings] = None,
        ldap_client_settings: Optional[OverhaveLdapClientSettings] = None,
        processor_settings: Optional[ProcessorSettings] = None,
        project_settings: Optional[OverhaveProjectSettings] = None,
        stash_client_settings: Optional[OverhaveStashClientSettings] = None,
        stash_project_settings: Optional[OverhaveStashProjectSettings] = None,
        test_settings: Optional[OverhaveTestSettings] = None,
        logging_settings: Optional[OverhaveLoggingSettings] = None,
        description_manager_settings: Optional[OverhaveDescriptionManagerSettings] = None,
    ) -> None:
        self.admin_settings = admin_settings or OverhaveAdminSettings()
        self.auth_settings = auth_settings or OverhaveAuthorizationSettings()
        self.compilation_settings = compilation_settings or OverhaveScenarioCompilerSettings()
        self.emulation_settings = emulation_settings or OverhaveEmulationSettings()
        self.file_settings = file_settings or OverhaveFileSettings()
        self.language_settings = language_settings or OverhaveLanguageSettings()
        self.processor_settings = processor_settings or ProcessorSettings()
        self.project_settings = project_settings or OverhaveProjectSettings()
        self.stash_client_settings = stash_client_settings or OverhaveStashClientSettings()
        self.stash_project_settings = stash_project_settings or OverhaveStashProjectSettings()
        self.test_settings = test_settings or OverhaveTestSettings()
        self.logging_settings = logging_settings or OverhaveLoggingSettings()
        self.description_manager_settings = description_manager_settings or OverhaveDescriptionManagerSettings()

        if self.auth_settings.auth_strategy is AuthorizationStrategy.LDAP:
            self.ldap_client_settings = ldap_client_settings or OverhaveLdapClientSettings()



class ScenarioCompilerError(Exception):
    """ Base exception for :class:`ScenarioCompiler` errors. """

    pass


class IncorrectScenarioTextError(ScenarioCompilerError):
    """ Exception for incorrect scenario text format. """

    pass


class ScenarioParserError(Exception):
    """ Base exception for parsing error. """

    pass


class FeatureNameParsingError(ScenarioParserError):
    """ Exception for feature name parsing error. """

    pass


class FeatureTypeParsingError(ScenarioParserError):
    """ Exception for feature type parsing error. """

    pass


class LastEditorParsingError(ScenarioParserError):
    """ Exception for last editor parsing error. """

    pass



import jinja2
from flask import Flask


def get_flask_app(template_folder: str) -> Flask:
    app = Flask('admin', template_folder=template_folder)
    app.secret_key = 'overhave cool secret key'
    app.config.update({'FLASK_ADMIN_SWATCH': 'united', 'FLASK_ADMIN_FLUID_LAYOUT': True})

    login_template_loader = jinja2.FileSystemLoader([template_folder])
    app.jinja_loader = jinja2.ChoiceLoader([app.jinja_loader, login_template_loader])  # type: ignore
    return app



class StashValidationError(ValueError):
    """ Exception for stash response validation error. """


class StashPrCreationError(RuntimeError):
    """ Exception for pull-request creation error. """


class NotSpecifiedFeatureTypeError(RuntimeError):
    """ Exception for incorrect feature type error. """



# flake8: noqa
from .abstract_factory import IOverhaveFactory
from .consumer_factory import ConsumerFactory
from .context import OverhaveContext
from .proxy_factory import ProxyFactory
from .proxy_getter import OverhaveFactoryType, get_proxy_factory



from overhave import db
from overhave.stash.models import StashPrCreationResponse


def add_pr_url(draft_id: int, response: StashPrCreationResponse) -> None:
    with db.create_session() as session:
        draft: db.Draft = session.query(db.Draft).filter(db.Draft.id == draft_id).one()
        draft.pr_url = response.get_pr_url()
        draft.created_at = response.created_date
        feature: db.Feature = session.query(db.Feature).filter(db.Feature.id == draft.feature.id).one()
        feature.released = response.open


def get_last_pr_url(feature_id: int) -> str:
    with db.create_session() as session:
        draft: db.Draft = session.query(db.Draft).filter(db.Draft.feature_id == feature_id).order_by(
            db.Draft.id.desc()
        ).first()
        return str(draft.pr_url)



# flake8: noqa
from .authorization import IAdminAuthorizationManager
from .converters import (
    EmulationRunModel,
    FeatureModel,
    FeatureTypeModel,
    ProcessingContext,
    ScenarioModel,
    TestRunModel,
)
from .emulator import Emulator
from .feature import (
    FeatureExtractor,
    FeatureTypeExtractionError,
    FeatureTypeName,
    IFeatureExtractor,
    ScenariosTestFileNotFound,
)
from .language import StepPrefixesModel, TranslitPack
from .settings import (
    OverhaveEmulationSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveRedisSettings,
    OverhaveScenarioCompilerSettings,
    ProcessorSettings,
)



import logging
from typing import Optional, cast

from wtforms import PasswordField, StringField

from overhave import db
from overhave.entities.authorization.manager.secret_mixin import AdminSecretMixin

logger = logging.getLogger(__name__)


class DefaultAdminAuthorizationManager(AdminSecretMixin):
    """ Class for user authorization via database.

    Manager authorize users with registered credentials from database.
    """

    def authorize_user(self, username_field: StringField, password_field: PasswordField) -> Optional[db.BaseUser]:
        with db.create_session(expire_on_commit=False) as s:
            db_user = (
                s.query(db.UserRole)
                .filter(db.UserRole.login == username_field.data, db.UserRole.password == password_field.data)
                .one_or_none()
            )
            if db_user is not None:
                return cast(db.BaseUser, db_user)
        return None



import abc
import enum
import json
from typing import Any, Dict, TypeVar, Union, cast

from pydantic.main import BaseModel


class RedisStream(str, enum.Enum):
    """ Enum that declares Redis streams. """

    TEST = "test-stream"
    EMULATION = "emulation-stream"

    @property
    def with_dunder(self) -> str:
        return cast(str, self.value.replace("-", "_"))


class _IRedisTask(BaseModel, abc.ABC):
    @property
    @abc.abstractmethod
    def message(self) -> Dict[bytes, bytes]:
        pass


class BaseRedisTask(_IRedisTask):
    """ Base task for Redis streams. """

    data: Any

    @property
    def message(self) -> Dict[bytes, bytes]:
        return {b'data': self.data.json().encode('utf-8')}


class TestRunData(BaseModel):
    """ Specific data for test run. """

    test_run_id: int


class TestRunTask(BaseRedisTask):
    """ Redis stream task for test run. """

    data: TestRunData


class EmulationData(BaseModel):
    """ Specific data for emulation run. """

    emulation_run_id: int


class EmulationTask(BaseRedisTask):
    """ Redis stream task for emulation run. """

    data: EmulationData


TRedisTask = TypeVar('TRedisTask', TestRunTask, EmulationTask, covariant=True)


class RedisPendingData(BaseModel):
    """ Class that describes pending data from Redis stream. """

    message_id: str
    consumer: str
    time_since_delivered: int
    times_delivered: int


class RedisUnreadData:
    """ Class for unread data from Redis stream. """

    def __init__(self, message_id: bytes, message: Dict[bytes, bytes]) -> None:
        self.message_id = message_id.decode()
        self.message = message

    @property
    def decoded_message(self) -> Dict[str, Any]:
        return {key.decode(): json.loads(value.decode('utf-8')) for key, value in self.message.items()}


class RedisContainer(BaseModel):
    """ Class for parsing instance of :class:`RedisUnreadData` from Redis stream.

    ```task``` will be parsed to one of declared models.
    """

    task: Union[TestRunTask, EmulationTask]



import logging
import typing
from pathlib import Path

import werkzeug
from flask import Flask, Response, redirect, send_from_directory

from overhave import db
from overhave.admin.flask import get_flask_admin, get_flask_app
from overhave.admin.flask.login_manager import get_flask_login_manager
from overhave.base_settings import DataBaseSettings
from overhave.factory import IOverhaveFactory, ProxyFactory

logger = logging.getLogger(__name__)

OverhaveAppType = typing.NewType('OverhaveAppType', Flask)


def _prepare_factory(factory: ProxyFactory) -> None:
    """ Resolve necessary settings and prepare instance of :class:`ProxyFactory` for usage. """
    DataBaseSettings().setup_db()
    factory.context.logging_settings.setup_logging()
    factory.patch_pytest()
    factory.supply_injector_for_collection()
    factory.injector.collect_configs()


def _resolved_app(factory: IOverhaveFactory, template_dir: Path) -> Flask:
    """ Resolve Flask application with :class:`IOverhaveFactory` and templates directory `template_dir`. """
    from overhave.admin.views import (
        DraftView,
        EmulationRunView,
        EmulationView,
        FeatureView,
        GroupView,
        OverhaveIndexView,
        TestingUserView,
        TestRunView,
        UserView,
    )

    index_view = OverhaveIndexView(name='Info', url='/', context=factory.context, auth_manager=factory.auth_manager)
    flask_admin = get_flask_admin(index_view=index_view)
    flask_admin.add_views(
        FeatureView(db.Feature, db.current_session, category="Scenarios", name="Features"),
        TestRunView(db.TestRun, db.current_session, category="Scenarios", name="Test runs"),
        DraftView(db.Draft, db.current_session, category="Scenarios", name="Versions"),
        EmulationView(db.Emulation, db.current_session, category="Emulation", name="Templates"),
        TestingUserView(db.TestUser, db.current_session, category="Emulation", name="Test users"),
        EmulationRunView(db.EmulationRun, db.current_session, category="Emulation", name="Emulation runs"),
        UserView(db.UserRole, db.current_session, category="Access", name="Users"),
        GroupView(db.GroupRole, db.current_session, category="Access", name="Groups"),
    )
    flask_app = get_flask_app(template_folder=template_dir.as_posix())
    flask_admin.init_app(app=flask_app)
    login_manager = get_flask_login_manager()
    login_manager.init_app(flask_app)
    db.ensure_feature_types_exist(factory.feature_extractor.feature_types)  # type: ignore
    return flask_app


def overhave_app(factory: ProxyFactory) -> OverhaveAppType:
    """ Overhave application, based on Flask. """
    current_dir = Path(__file__).parent
    template_dir = current_dir / 'templates'
    files_dir = current_dir / 'files'

    _prepare_factory(factory)
    flask_app = _resolved_app(factory=factory, template_dir=template_dir)
    flask_app.config["FILES_DIR"] = files_dir

    @flask_app.teardown_request
    def remove_session(exception: typing.Optional[Exception]) -> None:
        db.current_session.remove()

    @flask_app.route('/reports/<path:report>')
    def get_report(report: str) -> Response:
        return typing.cast(Response, send_from_directory(factory.context.file_settings.tmp_reports_dir, report))

    @flask_app.route('/emulations/<path:url>')
    def go_to_emulation(url: str) -> werkzeug.Response:
        return redirect(factory.context.emulation_settings.get_emulation_url(url))

    @flask_app.route('/pull_request/<int:run_id>')
    def create_pr(run_id: int) -> werkzeug.Response:
        return factory.processor.create_pull_request(run_id)

    @flask_app.route('/files/<path:file>')
    def get_files(file: str) -> Response:
        return typing.cast(Response, send_from_directory(files_dir, file))

    @flask_app.route('/favicon.ico')
    def favicon() -> Response:
        return typing.cast(Response, send_from_directory(files_dir, 'favicon.ico', mimetype='image/vnd.microsoft.icon'))

    return OverhaveAppType(flask_app)



from functools import cached_property
from typing import Callable, Dict, Type

from overhave.base_settings import DataBaseSettings
from overhave.factory import IOverhaveFactory
from overhave.redis.consumer import RedisConsumer
from overhave.redis.objects import BaseRedisTask, EmulationTask, RedisStream, TestRunTask, TRedisTask
from overhave.redis.runner import RedisConsumerRunner


class ConsumerFactory:
    """ Factory for :class:`RedisConsumer`, :class:`RedisConsumerRunner` and tasks mapping. """

    def __init__(self, factory: IOverhaveFactory, stream: RedisStream):
        self._factory = factory
        self._stream = stream
        DataBaseSettings().setup_db()

    @cached_property
    def _consumer(self) -> RedisConsumer:
        from overhave.entities import OverhaveRedisSettings

        return RedisConsumer(settings=OverhaveRedisSettings(), stream_name=self._stream)

    @cached_property
    def _mapping(self) -> Dict[Type[BaseRedisTask], Callable[[TRedisTask], None]]:
        return {
            EmulationTask: self._factory.emulator.start_emulation,  # type: ignore
            TestRunTask: lambda _: None,
        }

    @cached_property
    def runner(self) -> RedisConsumerRunner:
        return RedisConsumerRunner(consumer=self._consumer, mapping=self._mapping)



import logging
from typing import Any

import requests.auth
import tenacity
from pydantic import ValidationError

from overhave.stash.errors import StashValidationError
from overhave.stash.models import STASH_RESPONSE_MODELS, AnyStashResponseModel, StashPrRequest
from overhave.stash.settings import OverhaveStashClientSettings

logger = logging.getLogger(__name__)


class _BearerAuth(requests.auth.AuthBase):
    def __init__(self, token: str):
        super().__init__()
        self.token = token

    def __call__(self, r: requests.PreparedRequest) -> requests.PreparedRequest:
        r.headers['Authorization'] = f'Bearer {self.token}'
        return r

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, _BearerAuth):
            return False
        return self.token == other.token


class StashClient:
    """ Class for communcation with remote Bitbucket server. """

    def __init__(self, settings: OverhaveStashClientSettings):
        self._settings = settings

    @tenacity.retry(
        reraise=True,
        retry=tenacity.retry_if_exception_type(requests.ConnectionError),
        stop=tenacity.stop_after_attempt(3),
        before_sleep=tenacity.before_sleep_log(logger, logger.level),
        after=tenacity.after_log(logger, logger.level),
    )
    def send_pr(self, pull_request: StashPrRequest) -> AnyStashResponseModel:
        url = self._settings.get_pr_url(
            project_key=pull_request.target_branch.repository.project.key,
            repository_name=pull_request.target_branch.repository.name,
        )
        response = requests.post(
            url.human_repr(), json=pull_request.dict(by_alias=True), auth=_BearerAuth(self._settings.auth_token)
        )
        data = response.json()
        for model in STASH_RESPONSE_MODELS:
            try:
                logger.debug("Trying to parse '%s'...", model)
                return model.parse_obj(data)  # type: ignore
            except ValidationError:
                logger.debug("Could not convert response to '%s'!", model, exc_info=True)
        raise StashValidationError(
            f"Could not parse Stash response while trying to create pull-request!\nResponse: {data}"
        )



from typing import Optional, cast

from overhave import db
from overhave.utils.time import get_current_time


def create_test_run(scenario_id: int, executed_by: str) -> int:
    with db.create_session() as session:
        scenario = cast(db.Scenario, session.query(db.Scenario).filter(db.Scenario.id == scenario_id).one())
        feature = cast(db.Feature, session.query(db.Feature).filter(db.Feature.id == scenario.feature_id).one())
        run = db.TestRun(  # type: ignore
            scenario_id=scenario_id,
            name=feature.name,
            start=get_current_time(),
            status=db.TestRunStatus.STARTED,
            report_status=db.TestReportStatus.EMPTY,
            executed_by=executed_by,
        )
        session.add(run)
        session.flush()
        return cast(int, run.id)


def set_run_status(run_id: int, status: db.TestRunStatus) -> None:
    with db.create_session() as session:
        run = cast(db.TestRun, session.query(db.TestRun).filter(db.TestRun.id == run_id).one())
        run.status = status
        if status.finished:
            run.end = get_current_time()


def save_draft(run_id: int) -> int:
    with db.create_session() as session:
        draft = session.query(db.Draft).as_unique(test_run_id=run_id)
        session.add(draft)
        session.flush()
        return cast(int, draft.id)


def set_report(run_id: int, status: db.TestReportStatus, report: Optional[str] = None) -> None:
    with db.create_session() as session:
        run = cast(db.TestRun, session.query(db.TestRun).filter(db.TestRun.id == run_id).one())
        run.report_status = status
        if isinstance(report, str):
            run.report = report


def set_traceback(run_id: int, traceback: str) -> None:
    with db.create_session() as session:
        run = cast(db.TestRun, session.query(db.TestRun).filter(db.TestRun.id == run_id).one())
        run.traceback = traceback



import click

from overhave import overhave_app
from overhave.cli.group import overhave
from overhave.factory import get_proxy_factory


def _run_admin(port: int, debug: bool) -> None:
    factory = get_proxy_factory()
    overhave_app(factory).run(host='0.0.0.0', port=port, debug=debug)


@overhave.command(short_help='Run Overhave web-service')
@click.option('--port', default=8076)
@click.option('--debug', default=False)
def admin(port: int, debug: bool) -> None:
    _run_admin(port, debug)



# flake8: noqa
from .consumer import RedisConsumer
from .objects import BaseRedisTask, EmulationTask, RedisStream, TestRunTask, TRedisTask
from .producer import RedisProducer
from .runner import RedisConsumerRunner



import logging
import subprocess
import tempfile
import uuid
from functools import cached_property
from multiprocessing.pool import ThreadPool
from os import makedirs
from pathlib import Path
from time import sleep
from typing import Optional

import flask
import werkzeug

from overhave import db
from overhave.entities.converters import ProcessingContext, get_context_by_test_run_id
from overhave.entities.settings import OverhaveFileSettings, ProcessorSettings
from overhave.processing.abstract import IProcessor
from overhave.scenario import FileManager
from overhave.stash import IStashProjectManager
from overhave.storage import save_draft, set_report, set_run_status, set_traceback
from overhave.storage.pull_request import add_pr_url
from overhave.testing import ConfigInjector, PytestRunner

logger = logging.getLogger(__name__)


class Processor(IProcessor):
    """ Class for application entity tasks processing in instance of :class:`ThreadPool`. """

    def __init__(
        self,
        settings: ProcessorSettings,
        file_settings: OverhaveFileSettings,
        injector: ConfigInjector,
        file_manager: FileManager,
        test_runner: PytestRunner,
        stash_manager: IStashProjectManager,
    ):
        self._settings = settings
        self._file_settings = file_settings
        self._injector = injector
        self._file_manager = file_manager
        self._test_runner = test_runner
        self._stash_manager = stash_manager

    @cached_property
    def _thread_pool(self) -> ThreadPool:
        return ThreadPool(processes=self._settings.processes_num)

    def _run_test(self, context: ProcessingContext, alluredir: Path) -> int:
        with self._file_manager.tmp_feature_file(context=context) as feature_file:
            with self._file_manager.tmp_fixture_file(context=context, feature_file=feature_file) as fixture_file:
                return self._test_runner.run(fixture_file=fixture_file.name, alluredir=alluredir.as_posix(),)

    def _generate_report(self, alluredir: Path, report_dir: Path) -> Optional[int]:
        generation_cmd = [
            self._settings.allure_cmdline,
            "generate",
            f"{alluredir}/",
            "--output",
            report_dir.as_posix(),
            "--clean",
        ]
        logger.debug("Allure report generation command: %s", " ".join(generation_cmd))
        makedirs(report_dir.as_posix(), exist_ok=True)
        try:
            return subprocess.run(
                generation_cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=self._settings.report_creation_timeout,
                check=True,
            ).returncode
        except (FileNotFoundError, subprocess.CalledProcessError):
            logger.exception("Error while generating Allure report!")
            return None

    def _process_run(self, run_id: int) -> None:
        set_run_status(run_id=run_id, status=db.TestRunStatus.RUNNING)
        try:
            ctx = get_context_by_test_run_id(run_id)

            allure_dir = Path(tempfile.mkdtemp())
            logger.debug("Allure results directory path: %s", allure_dir.as_posix())

            test_return_code = self._run_test(context=ctx, alluredir=allure_dir)
            logger.debug("Test returncode: %s", test_return_code)

            if test_return_code == 0:
                set_run_status(run_id=run_id, status=db.TestRunStatus.SUCCESS)
            else:
                set_traceback(run_id, "Test run failed! Something went wrong...")
                set_run_status(run_id=run_id, status=db.TestRunStatus.FAILED)

            allure_tmpdir_name = str(uuid.uuid1())
            logger.debug("Allure report directory: %s", allure_tmpdir_name)
            report_dir = self._file_settings.tmp_reports_dir / allure_tmpdir_name

            report_generation_returncode = self._generate_report(alluredir=allure_dir, report_dir=report_dir)
            if report_generation_returncode == 0:
                logger.debug(
                    "Allure report successfully generated to directory: %s", report_dir.as_posix(),
                )
                set_report(
                    run_id=run_id, status=db.TestReportStatus.GENERATED, report=f"{allure_tmpdir_name}/index.html"
                )
            else:
                set_report(run_id=run_id, status=db.TestReportStatus.GENERATION_FAILED)

        except Exception as e:
            logger.exception("Error!")
            set_run_status(run_id, db.TestRunStatus.INTERNAL_ERROR)
            set_traceback(run_id, f"{e}")

    def execute_test(self, test_run_id: int) -> werkzeug.Response:
        self._thread_pool.apply_async(self._process_run, args=(test_run_id,))
        logger.debug("Redirect to TestRun details view with run_id %s", test_run_id)
        return flask.redirect(flask.url_for('testrun.details_view', id=test_run_id))

    def _create_pull_request(self, test_run_id: int) -> None:
        try:
            draft_id = save_draft(test_run_id)
            response = self._stash_manager.create_pull_request(test_run_id)
            logger.info("Stash PR response has been gotten")
            add_pr_url(draft_id=draft_id, response=response)
            logger.info("Stash PR result has been written")
        except Exception:
            logger.exception('Error while trying to create Stash PR!')

    def create_pull_request(self, test_run_id: int) -> werkzeug.Response:
        self._thread_pool.apply_async(self._create_pull_request, args=(test_run_id,))
        sleep(3)
        return flask.redirect(flask.url_for("draft.details_view"))



from flask_login import current_user
from wtforms import ValidationError

from overhave.admin.views.base import ModelViewProtected
from overhave.db import Role


class DraftView(ModelViewProtected):
    """ View for :class:`Draft` table. """

    def on_model_delete(self, model) -> None:  # type: ignore
        if not current_user.role == Role.admin:
            raise ValidationError('Only administrator could delete published version of scenario!')

    can_delete = True
    column_list = ('id', 'feature_id', 'test_run_id', 'pr_url', 'created_at')
    column_exclude_list = ('feature', 'text')  # type: ignore



from typing import List, Mapping, Optional

from pydantic import validator
from pydantic.datetime_parse import timedelta
from yarl import URL

from overhave.base_settings import BaseOverhavePrefix
from overhave.entities.feature import FeatureTypeName
from overhave.stash.errors import NotSpecifiedFeatureTypeError
from overhave.stash.models import StashBranch, StashProject, StashRepository, StashReviewer, StashReviewerInfo


class OverhaveStashProjectSettings(BaseOverhavePrefix):
    """ Settings for :class:`StashProjectManager`.

    This is a representation of BitBucket project parameters.
    Some pull-request parameters are also could be defined through these settings.
    """

    repository_name: str  # for example 'bdd-features'
    project_key: str  # for example 'PRJ'
    default_target_branch_name: str = 'master'

    # Pull-request default reviewers as list
    default_reviewers: List[str] = []
    # Pull-request default reviewers as mapping with :class:```FeatureTypeName```
    feature_type_to_reviewers_mapping: Mapping[FeatureTypeName, List[str]] = {}

    @property
    def repository(self) -> StashRepository:
        return StashRepository(slug=self.repository_name, project=StashProject(key=self.project_key))

    @property
    def target_branch(self) -> StashBranch:
        return StashBranch(id=self.default_target_branch_name, repository=self.repository)

    def get_reviewers(self, feature_type: FeatureTypeName) -> List[StashReviewer]:
        if self.feature_type_to_reviewers_mapping:
            reviewers = self.feature_type_to_reviewers_mapping[feature_type]
            if not reviewers:
                raise NotSpecifiedFeatureTypeError(
                    f"FeatureTypeName '{feature_type}' key was not specified in "
                    "'feature_type_to_reviewers_mapping' dict!"
                )
        else:
            reviewers = self.default_reviewers
        return [StashReviewer(user=StashReviewerInfo(name=reviewer)) for reviewer in reviewers]


class OverhaveStashClientSettings(BaseOverhavePrefix):
    """ Settings for :class:`StashClient`. """

    stash_url: URL  # for example "https://my.company"
    pr_path: str = "rest/api/1.0/projects/{project_key}/repos/{repository_name}/pull-requests"

    connect_timeout: timedelta = timedelta(seconds=5)
    read_timeout: timedelta = timedelta(seconds=10)

    auth_token: str

    @validator('stash_url', pre=True)
    def make_url(cls, v: Optional[str]) -> Optional[URL]:
        if v is not None and isinstance(v, str):
            return URL(v)
        return v

    def get_pr_url(self, project_key: str, repository_name: str) -> URL:
        return self.stash_url / self.pr_path.format(project_key=project_key, repository_name=repository_name)



import abc

import werkzeug


class IProcessor(abc.ABC):
    """ Abstract class for application processor. """

    @abc.abstractmethod
    def execute_test(self, test_run_id: int) -> werkzeug.Response:
        pass

    @abc.abstractmethod
    def create_pull_request(self, test_run_id: int) -> werkzeug.Response:
        pass



# flake8: noqa
from .access import GroupView, UserView
from .draft import DraftView
from .emulation import EmulationView
from .emulation_run import EmulationRunView
from .feature import FeatureView
from .formatters import datetime_formatter, result_report_formatter, task_formatter
from .index import OverhaveIndexView
from .scenario_test_run import TestRunView
from .testing_users import TestingUserView



# flake8: noqa
from .client import StashClient
from .manager import IStashProjectManager
from .models import StashBranch, StashProject, StashRepository, StashReviewer, StashReviewerInfo
from .settings import OverhaveStashClientSettings, OverhaveStashProjectSettings



from functools import lru_cache
from typing import NewType

from overhave.factory.proxy_factory import ProxyFactory

OverhaveFactoryType = NewType('OverhaveFactoryType', ProxyFactory)


@lru_cache(maxsize=None)
def get_proxy_factory() -> OverhaveFactoryType:
    return OverhaveFactoryType(ProxyFactory())



# flake8: noqa
from .config_injector import ConfigInjector
from .plugin_utils import (
    add_issue_links_to_report,
    add_scenario_title_to_report,
    get_description_manager,
    get_scenario,
    has_issue_links,
    is_pytest_bdd_item,
    set_issue_links,
)
from .settings import (
    EmptyBrowseURLError,
    OverhaveDescriptionManagerSettings,
    OverhaveProjectSettings,
    OverhaveTestSettings,
)
from .step_collector import StepCollector
from .test_runner import PytestRunner



from __future__ import annotations

from typing import Optional, cast

import sqlalchemy as sa
import sqlalchemy_utils as su
from flask import url_for
from sqlalchemy import orm as so

from overhave.db.base import Base, PrimaryKeyMixin, PrimaryKeyWithoutDateMixin
from overhave.db.statuses import EmulationStatus, TestReportStatus, TestRunStatus
from overhave.db.types import ARRAY_TYPE, DATETIME_TYPE, INT_TYPE, LONG_STR_TYPE, SHORT_STR_TYPE, TEXT_TYPE
from overhave.db.users import UserRole


class FeatureType(Base, PrimaryKeyWithoutDateMixin):
    """ Feature types table. """

    name = sa.Column(sa.Text, unique=True, nullable=False, doc='Feature types choice')

    def __repr__(self) -> str:
        return cast(str, self.name.upper())


@su.generic_repr('name', 'last_edited_by')
class Feature(Base, PrimaryKeyMixin):
    """ Features table. """

    name = sa.Column(LONG_STR_TYPE, doc="Feature name", nullable=False, unique=True)
    author = sa.Column(
        SHORT_STR_TYPE, sa.ForeignKey(UserRole.login), doc="Feature author login", nullable=False, index=True
    )
    type_id = sa.Column(INT_TYPE, sa.ForeignKey(FeatureType.id), nullable=False, doc='Feature types choice')
    task = sa.Column(ARRAY_TYPE, doc="Feature tasks list", nullable=False)
    last_edited_by = sa.Column(SHORT_STR_TYPE, doc="Last feature editor login", nullable=False)
    released = sa.Column(sa.Boolean, doc="Feature release state boolean", nullable=False, default=False)

    feature_type = so.relationship(FeatureType)


@su.generic_repr('feature_id')
class Scenario(Base, PrimaryKeyMixin):
    """ Scenarios table. """

    feature_id = sa.Column(INT_TYPE, sa.ForeignKey(Feature.id), nullable=False, unique=True)
    text = sa.Column(TEXT_TYPE, doc="Text storage for scenarios in feature")

    feature = so.relationship(Feature, backref=so.backref("scenario", cascade="all, delete-orphan"))


class TestRun(Base, PrimaryKeyMixin):
    """ Test runs table. """

    scenario_id = sa.Column(INT_TYPE, sa.ForeignKey(Scenario.id), nullable=False, index=True)
    name = sa.Column(LONG_STR_TYPE, nullable=False)
    start = sa.Column(DATETIME_TYPE, doc="Test start time", nullable=False)
    end = sa.Column(DATETIME_TYPE, doc="Test finish time")
    status = sa.Column(sa.Enum(TestRunStatus), doc="Current test status", nullable=False)
    report_status = sa.Column(sa.Enum(TestReportStatus), doc="Report generation result", nullable=False)
    executed_by = sa.Column(SHORT_STR_TYPE, sa.ForeignKey(UserRole.login), doc="Test executor login", nullable=False)
    report = sa.Column(TEXT_TYPE, doc="Relative report URL")
    traceback = sa.Column(TEXT_TYPE, doc="Text storage for error traceback")

    scenario = so.relationship(Scenario, backref=so.backref("test_runs", cascade="all, delete-orphan"))


class DraftQuery(so.Query):
    """ Scenario versions table. """

    def as_unique(self, test_run_id: int) -> Draft:
        with self.session.no_autoflush:
            run = self.session.query(TestRun).get(test_run_id)
            draft: Optional[Draft] = self.session.query(Draft).filter(
                Draft.test_run_id == run.id, Draft.text == run.scenario.text
            ).one_or_none()

        if draft:
            return draft

        return Draft(  # type: ignore
            test_run_id=test_run_id, feature_id=run.scenario.feature_id, text=run.scenario.text
        )


class Draft(Base, PrimaryKeyMixin):
    """ Scenario versions table. """

    __query_cls__ = DraftQuery

    feature_id = sa.Column(INT_TYPE, sa.ForeignKey(Feature.id), nullable=False, index=True)
    test_run_id = sa.Column(INT_TYPE, sa.ForeignKey(TestRun.id), nullable=False)
    text = sa.Column(TEXT_TYPE, doc="Released scenario text")
    pr_url = sa.Column(sa.Text, doc="Absolute pull-request URL", nullable=True)

    feature = so.relationship(Feature, backref=so.backref("versions", cascade="all, delete-orphan"))

    __table_args__ = (sa.UniqueConstraint(test_run_id),)

    def __html__(self) -> str:
        return f'<a href="{url_for("draft.details_view", id=self.id)}">Draft: {self.id}</a>'


@su.generic_repr('id', 'name', 'created_by')
class TestUser(Base, PrimaryKeyMixin):
    """ Test users table. """

    name = sa.Column(LONG_STR_TYPE, nullable=False, unique=True)
    feature_type_id = sa.Column(INT_TYPE, sa.ForeignKey(FeatureType.id), nullable=False, doc='Feature types choice')
    specification = sa.Column(sa.JSON(none_as_null=True))
    created_by = sa.Column(SHORT_STR_TYPE, sa.ForeignKey(UserRole.login), doc="Author login", nullable=False)

    feature_type = so.relationship(FeatureType)


class Emulation(Base, PrimaryKeyMixin):
    """ Emulation templates table. """

    name = sa.Column(LONG_STR_TYPE, nullable=False, unique=True)
    test_user_id = sa.Column(INT_TYPE, sa.ForeignKey(TestUser.id), nullable=False, doc='Test user ID')
    command = sa.Column(TEXT_TYPE, nullable=False, doc="Command for emulator's execution")
    created_by = sa.Column(SHORT_STR_TYPE, sa.ForeignKey(UserRole.login), doc="Author login", nullable=False)
    test_user = so.relationship(TestUser)


class EmulationRun(Base, PrimaryKeyMixin):
    """ Emulation runs table. """

    __tablename__ = 'emulation_run'  # type: ignore
    emulation_id = sa.Column(INT_TYPE, sa.ForeignKey(Emulation.id), nullable=False, index=True)
    status = sa.Column(sa.Enum(EmulationStatus), doc="Current emulation status", nullable=False)
    changed_at = sa.Column(sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now())
    port = sa.Column(INT_TYPE, doc="Port for emulation")
    traceback = sa.Column(TEXT_TYPE, doc="Text storage for error traceback")
    initiated_by = sa.Column(
        SHORT_STR_TYPE, sa.ForeignKey(UserRole.login), doc="Initiator of start emulation", nullable=False
    )

    emulation = so.relationship(Emulation, backref=so.backref("emulation_runs", cascade="all, delete-orphan"))

    def __init__(self, emulation_id: int, initiated_by: str):
        self.emulation_id = emulation_id
        self.status = EmulationStatus.CREATED
        self.initiated_by = initiated_by



import enum


class TestRunStatus(str, enum.Enum):
    """ Enum for test run statuses. """

    STARTED = 'STARTED'
    RUNNING = 'RUNNING'
    FAILED = 'FAILED'
    SUCCESS = 'SUCCESS'
    INTERNAL_ERROR = 'INTERNAL_ERROR'

    @property
    def finished(self) -> bool:
        return self in (TestRunStatus.FAILED, TestRunStatus.SUCCESS, TestRunStatus.INTERNAL_ERROR)


class TestReportStatus(str, enum.Enum):
    """ Enum for test run statuses. """

    EMPTY = 'EMPTY'
    GENERATION_FAILED = 'GENERATION_FAILED'
    GENERATED = 'GENERATED'
    SAVED = 'SAVED'

    @property
    def has_report(self) -> bool:
        return self in (TestReportStatus.GENERATED, TestReportStatus.SAVED)


class EmulationStatus(str, enum.Enum):
    """ Enum for emulation statuses. """

    CREATED = 'CREATED'
    REQUESTED = 'REQUESTED'
    READY = 'READY'
    ERROR = 'ERROR'

    @property
    def processed(self) -> bool:
        return self in (EmulationStatus.READY, EmulationStatus.ERROR)



from functools import cached_property
from os import makedirs
from typing import Any, Dict, Optional

from overhave.entities import FeatureExtractor, IFeatureExtractor, OverhaveRedisSettings
from overhave.entities.authorization.manager import IAdminAuthorizationManager, LDAPAuthenticator
from overhave.entities.authorization.mapping import AUTH_STRATEGY_TO_MANAGER_MAPPING, AuthorizationStrategy
from overhave.entities.emulator import EmulationTask, Emulator
from overhave.factory.abstract_factory import IOverhaveFactory
from overhave.factory.context import OverhaveContext
from overhave.processing import IProcessor
from overhave.redis import RedisProducer, RedisStream
from overhave.scenario import FileManager, ScenarioCompiler, ScenarioParser
from overhave.stash import IStashProjectManager, StashClient
from overhave.storage import EmulationStorage, FeatureTypeStorage, IEmulationStorage, IFeatureTypeStorage
from overhave.testing import ConfigInjector, PytestRunner, StepCollector


class OverhaveBaseFactory(IOverhaveFactory):
    """ Base factory for application entities resolution and usage. """

    def __init__(self) -> None:
        self._context: Optional[OverhaveContext] = None

    def set_context(self, context: OverhaveContext) -> None:
        self._context = context

    @property
    def has_context(self) -> bool:
        return self._context is not None

    @property
    def context(self) -> OverhaveContext:
        if self._context is None:
            self._context = OverhaveContext()
        return self._context

    @cached_property
    def _test_runner(self) -> PytestRunner:
        return PytestRunner(test_settings=self.context.test_settings)

    @property
    def test_runner(self) -> PytestRunner:
        return self._test_runner

    @property
    def step_collector(self) -> StepCollector:
        return StepCollector(step_prefixes=self.context.language_settings.step_prefixes)

    @cached_property
    def _injector(self) -> ConfigInjector:
        return ConfigInjector()

    @property
    def injector(self) -> ConfigInjector:
        return self._injector

    @cached_property
    def _feature_extractor(self) -> IFeatureExtractor:
        return FeatureExtractor(file_settings=self.context.file_settings)

    @property
    def feature_extractor(self) -> IFeatureExtractor:
        return self._feature_extractor

    @cached_property
    def _file_manager(self) -> FileManager:
        for path in (
            self.context.file_settings.tmp_features_dir,
            self.context.file_settings.tmp_fixtures_dir,
            self.context.file_settings.tmp_reports_dir,
        ):
            makedirs(path.as_posix(), exist_ok=True)
        return FileManager(
            project_settings=self.context.project_settings,
            file_settings=self.context.file_settings,
            language_settings=self.context.language_settings,
            feature_extractor=self._feature_extractor,
            scenario_compiler=ScenarioCompiler(
                compilation_settings=self.context.compilation_settings,
                language_settings=self.context.language_settings,
                task_links_keyword=self.context.project_settings.links_keyword,
            ),
            scenario_parser=ScenarioParser(
                compilation_settings=self.context.compilation_settings,
                language_settings=self.context.language_settings,
                feature_extractor=self._feature_extractor,
                task_links_keyword=self.context.project_settings.links_keyword,
            ),
        )

    @property
    def file_manager(self) -> FileManager:
        return self._file_manager

    @cached_property
    def _stash_client(self) -> StashClient:
        return StashClient(settings=self.context.stash_client_settings)

    @cached_property
    def _stash_manager(self) -> IStashProjectManager:
        from overhave.stash.manager.manager import StashProjectManager

        return StashProjectManager(
            stash_project_settings=self.context.stash_project_settings,
            file_settings=self.context.file_settings,
            client=self._stash_client,
            file_manager=self._file_manager,
            task_links_keyword=self.context.project_settings.links_keyword,
        )

    @cached_property
    def _processor(self) -> IProcessor:
        from overhave.processing.processor import Processor

        return Processor(
            settings=self.context.processor_settings,
            file_settings=self.context.file_settings,
            injector=self._injector,
            file_manager=self._file_manager,
            test_runner=self._test_runner,
            stash_manager=self._stash_manager,
        )

    @property
    def processor(self) -> IProcessor:
        return self._processor

    @cached_property
    def _auth_manager(self) -> IAdminAuthorizationManager:
        settings = self.context.auth_settings
        kwargs: Dict[str, Any] = dict(settings=settings)
        if settings.auth_strategy is AuthorizationStrategy.LDAP:
            kwargs["ldap_authenticator"] = LDAPAuthenticator(settings=self.context.ldap_client_settings)
        return AUTH_STRATEGY_TO_MANAGER_MAPPING[settings.auth_strategy](**kwargs)  # type: ignore

    @property
    def auth_manager(self) -> IAdminAuthorizationManager:
        return self._auth_manager

    @cached_property
    def _emulator(self) -> Emulator:
        return Emulator(storage=self._emulation_storage, settings=self.context.emulation_settings)

    @property
    def emulator(self) -> Emulator:
        return self._emulator

    @cached_property
    def _feature_type_storage(self) -> IFeatureTypeStorage:
        return FeatureTypeStorage()

    @property
    def feature_type_storage(self) -> IFeatureTypeStorage:
        return self._feature_type_storage

    @cached_property
    def _redis_producer(self) -> RedisProducer:
        return RedisProducer(settings=OverhaveRedisSettings(), mapping={EmulationTask: RedisStream.EMULATION})

    @property
    def redis_producer(self) -> RedisProducer:
        return self._redis_producer

    @cached_property
    def _emulation_storage(self) -> IEmulationStorage:
        return EmulationStorage(settings=self.context.emulation_settings)

    @property
    def emulation_storage(self) -> IEmulationStorage:
        return self._emulation_storage



import datetime
from typing import Any, List, Optional

from flask_admin.contrib.sqla import ModelView
from markupsafe import Markup

from overhave.db import TestReportStatus, TestRunStatus


def datetime_formatter(view: ModelView, context: Any, model: Any, name: str) -> Markup:
    time: Optional[datetime.datetime] = getattr(model, name)
    if time:
        return Markup(time.strftime('%d-%m-%Y %H:%M:%S'))
    return Markup("")


def task_formatter(view: ModelView, context: Any, model: Any, name: str) -> Markup:
    browse_url = getattr(view, 'browse_url')
    tasks = getattr(model, name)
    if not browse_url or not tasks:
        return Markup("")
    task_links: List[str] = []
    for task in tasks:
        task_links.append(f"<a href='{browse_url}/{task}' target='blank'>{task}</a>")
    return Markup(", ".join(task_links))


def _get_button_class_by_status(status: str) -> str:
    enum_member = TestRunStatus[status]
    if enum_member is TestRunStatus.SUCCESS:
        return "success-btn"
    return "default-btn"


def result_report_formatter(view: ModelView, context: Any, model: Any, name: str) -> Markup:
    status = getattr(model, name)
    if not status:
        return Markup("")

    report_status = TestReportStatus[getattr(model, 'report_status')]
    if report_status.has_report:
        action = f"href='/reports/{getattr(model, 'report')}' target='_blank'"
        title = "Go to report"
    else:
        action = f"href='/testrun/details/?id={model.id}'"
        title = "Show details"

    return Markup(
        f"<a {action}"
        f"<form action='#'>"
        f"<fieldset title='{title}'>"
        f"<button class='link-button {_get_button_class_by_status(status)}'>{status}</button>"
        "</fieldset>"
        "</form>"
        "</a>"
    )


def json_formatter(view: ModelView, context: Any, model: Any, name: str) -> Markup:
    data = getattr(model, name)
    if not data:
        return Markup("")
    info = ""
    for key, value in data.items():
        if value:
            info += f"<b>{key}</b>:&nbsp;&nbsp;{value}<br>"
    return Markup("<form>" "<fieldset>" f"<div class='json-data'>{info}</div>" "</fieldset>" "</form>")



import logging
from typing import Optional

import git

from overhave.entities.converters import ProcessingContext, get_context_by_test_run_id
from overhave.entities.settings import OverhaveFileSettings
from overhave.scenario import FileManager, generate_task_info
from overhave.stash.client import StashClient
from overhave.stash.errors import StashPrCreationError, StashValidationError
from overhave.stash.manager.abstract import IStashProjectManager
from overhave.stash.models import StashBranch, StashErrorResponse, StashPrCreationResponse, StashPrRequest
from overhave.stash.settings import OverhaveStashProjectSettings
from overhave.storage.pull_request import get_last_pr_url
from overhave.utils.time import get_current_time

logger = logging.getLogger(__name__)


class StashCommonMixin:
    """ Mixin for :class:`StashProjectManager`. """

    @staticmethod
    def _create_head(name: str, repository: git.Repo) -> git.Head:
        origin: git.Head = repository.remotes.origin
        try:
            origin.pull()
        except git.GitCommandError as e:
            raise RuntimeError from e

        logger.info("Origin successfully pulled")
        if name in repository.heads:
            repository.delete_head(repository.heads[name], force=True)
            logger.warning("Trashed existed branch with name '%s'", name)

        branch = repository.create_head(name)
        logger.info("Created head for branch: %s", branch)
        assert repository.active_branch != branch
        return branch

    @staticmethod
    def _make_compatible_str(input_str: str) -> str:
        return input_str.lower().replace(" ", "_")

    @staticmethod
    def _generate_response(
        title: str, pr_url: Optional[str] = None, traceback: Optional[Exception] = None
    ) -> StashPrCreationResponse:
        return StashPrCreationResponse(
            title=title,
            open=False,
            traceback=traceback,
            createdDate=get_current_time(),
            updatedDate=get_current_time(),
            pull_request_url=pr_url,
        )


class StashProjectManager(StashCommonMixin, IStashProjectManager):
    """ Class for feature pull requests management. """

    def __init__(
        self,
        stash_project_settings: OverhaveStashProjectSettings,
        file_settings: OverhaveFileSettings,
        client: StashClient,
        file_manager: FileManager,
        task_links_keyword: Optional[str],
    ):
        self._stash_project_settings = stash_project_settings
        self._file_settings = file_settings
        self._client = client
        self._file_manager = file_manager
        self._task_links_keyword = task_links_keyword

    def _make_description(self, context: ProcessingContext) -> str:
        return "\n".join(
            (
                f"Feature ID: {context.feature.id}. Type: '{context.feature.feature_type.name}'.",
                f"Created by: @{context.feature.author} at {context.feature.created_at}.",
                f"Last edited by: @{context.feature.last_edited_by}.",
                generate_task_info(tasks=context.feature.task, header=self._task_links_keyword),
                f"PR from Test Run ID: {context.test_run.id}. Executed by: @{context.test_run.executed_by}",
            )
        )

    def _create_commit(self, context: ProcessingContext, repository: git.Repo, branch: git.Head) -> None:
        original_branch = repository.active_branch
        changes_pushed = False
        try:
            branch.checkout()
            logger.info("Active branch after checkout: %s", repository.active_branch)
            # Раньше было self._path_settings.bdd_chat_features_dir
            feature_file = self._file_manager.produce_feature_file(context=context)
            logger.info("Untracked files after feature creation: %s", repository.untracked_files)
            repository.index.add([feature_file.as_posix()])
            logger.info("Untracked files after index update: %s", repository.untracked_files)

            repository.index.commit(
                (
                    f"Added feature {context.feature.name} "
                    f"(id {context.feature.id}) "
                    f"by {context.test_run.executed_by}"
                ),
                skip_hooks=True,
            )
            logger.info("Commit successfully created")
            assert (
                repository.active_branch != original_branch
            ), f"Current branch could not be equal to '{original_branch}'!"
            assert repository.active_branch == branch, f"Current branch should be equal to '{branch}'!"

            logger.info("Remote origin url: %s", repository.remotes.origin.url)
            repository.git.push("origin", branch, force=True)
            logger.info("Changes successfully pushed to remote repository")
            changes_pushed = True
        except (git.GitCommandError, AssertionError):
            logger.exception("Error while trying to commit feature file!")

        if original_branch != repository.active_branch:
            logger.info("Current branch is '%s'", repository.active_branch)
            original_branch.checkout()
            logger.info("Branch returned to '%s'", repository.active_branch)

        if not changes_pushed:
            raise RuntimeError("Commit has not been created!")

    def create_pull_request(self, test_run_id: int) -> StashPrCreationResponse:
        ctx = get_context_by_test_run_id(test_run_id)

        repository = git.Repo(self._file_settings.features_base_dir)
        logger.info("Repository: %s", repository)
        logger.info("Working dir: %s", repository.working_dir)
        logger.info("Active head: %s", repository.active_branch)

        branch_name = f"bdd-feature-{ctx.feature.id}"
        try:
            branch = self._create_head(name=branch_name, repository=repository)
            self._create_commit(context=ctx, repository=repository, branch=branch)
        except RuntimeError as e:
            logger.exception('Error while trying to commit code!')
            return self._generate_response(title=ctx.feature.name, traceback=e)

        pull_request = StashPrRequest(
            title=f"BDD {ctx.feature.name}",
            description=self._make_description(ctx),
            open=True,
            fromRef=StashBranch(id=branch_name, repository=self._stash_project_settings.repository),
            toRef=self._stash_project_settings.target_branch,
            reviewers=self._stash_project_settings.get_reviewers(feature_type=ctx.feature.feature_type.name),
        )
        logger.info('Prepared PR: %s', pull_request.json(by_alias=True))
        try:
            response = self._client.send_pr(pull_request)
            if isinstance(response, StashPrCreationResponse):
                return response
            if isinstance(response, StashErrorResponse) and response.duplicate:
                return self._generate_response(
                    title=pull_request.title or ctx.feature.name, pr_url=get_last_pr_url(feature_id=ctx.feature.id)
                )
            raise StashPrCreationError(response)
        except StashValidationError as e:
            logger.exception("PR has not been created in Stash repository!")
            fake_response = self._generate_response(title=pull_request.title, traceback=e)  # type: ignore
            logger.warning("Created fail response: %s", fake_response)
            return fake_response



from overhave.entities import Emulator, IAdminAuthorizationManager, IFeatureExtractor
from overhave.factory.abstract_factory import IOverhaveFactory
from overhave.factory.base_factory import OverhaveBaseFactory
from overhave.factory.context import OverhaveContext
from overhave.processing import IProcessor
from overhave.redis import RedisProducer
from overhave.scenario import FileManager
from overhave.storage import IEmulationStorage, IFeatureTypeStorage
from overhave.testing import ConfigInjector, PytestRunner


class ProxyFactory(IOverhaveFactory):
    """ Factory for application entities resolution and usage, based on proxy-object pattern.

    Class inherits :class:`IOverhaveFactory` and realise logic for dynamic entities resolution before and
    during application or tests start-up. In fact, it is a proxy-object for :class:`OverhaveBaseFactory`.
    """

    def __init__(self) -> None:
        self._factory = OverhaveBaseFactory()
        self._pytest_patched = False
        self._collection_prepared = False

    def set_context(self, context: OverhaveContext) -> None:
        self._factory.set_context(context)

    @property
    def context(self) -> OverhaveContext:
        return self._factory.context

    @property
    def test_runner(self) -> PytestRunner:
        return self._factory.test_runner

    @property
    def file_manager(self) -> FileManager:
        return self._factory.file_manager

    @property
    def injector(self) -> ConfigInjector:
        return self._factory.injector

    @property
    def feature_extractor(self) -> IFeatureExtractor:
        return self._factory.feature_extractor

    def patch_pytest(self) -> None:
        if not self._pytest_patched:
            self._factory.injector.patch_pytestbdd_prefixes(
                custom_step_prefixes=self._factory.context.language_settings.step_prefixes
            )
            self._pytest_patched = True

    @property
    def pytest_patched(self) -> bool:
        return self._pytest_patched

    def supply_injector_for_collection(self) -> None:
        if not self._collection_prepared:
            self._factory.injector.supplement_on_fly(
                project_settings=self._factory.context.project_settings,
                file_settings=self._factory.context.file_settings,
                step_collector=self._factory.step_collector,
                test_runner=self._factory.test_runner,
                feature_types=self._factory.feature_extractor.feature_types,
            )
            self._collection_prepared = True

    @property
    def processor(self) -> IProcessor:
        return self._factory.processor

    @property
    def auth_manager(self) -> IAdminAuthorizationManager:
        return self._factory.auth_manager

    @property
    def emulator(self) -> Emulator:
        return self._factory.emulator

    @property
    def feature_type_storage(self) -> IFeatureTypeStorage:
        return self._factory.feature_type_storage

    @property
    def redis_producer(self) -> RedisProducer:
        return self._factory.redis_producer

    @property
    def emulation_storage(self) -> IEmulationStorage:
        return self._factory.emulation_storage



from _pytest.nodes import Item
from pytest_bdd.parser import Scenario, Step


def get_scenario(item: Item) -> Scenario:
    return item._obj.__scenario__  # type: ignore


def is_pytest_bdd_item(item: Item) -> bool:
    if hasattr(item, "_obj"):
        return hasattr(item._obj, "__scenario__") and isinstance(get_scenario(item), Scenario)  # type: ignore
    return False


def get_full_step_name(step: Step) -> str:
    return f"{step.keyword} {step._name}"



from datetime import datetime
from typing import List

from pydantic.main import BaseModel
from pydantic_sqlalchemy import sqlalchemy_to_pydantic

from overhave.db import (
    Emulation,
    EmulationRun,
    Feature,
    FeatureType,
    Scenario,
    TestReportStatus,
    TestRun,
    TestRunStatus,
    TestUser,
    create_session,
)
from overhave.entities.feature import FeatureTypeName


class FeatureTypeModel(sqlalchemy_to_pydantic(FeatureType)):  # type: ignore
    """ Model for :class:`FeatureType` row. """

    name: FeatureTypeName


class FeatureModel(sqlalchemy_to_pydantic(Feature)):  # type: ignore
    """ Model for :class:`Feature` row. """

    name: str
    author: str
    feature_type: FeatureTypeModel
    last_edited_by: str
    task: List[str]


class ScenarioModel(sqlalchemy_to_pydantic(Scenario)):  # type: ignore
    """ Model for :class:`Scenario` row. """

    text: str


class TestRunModel(sqlalchemy_to_pydantic(TestRun)):  # type: ignore
    """ Model for :class:`TestRun` row. """

    id: int
    executed_by: str
    start: datetime
    status: TestRunStatus
    report_status: TestReportStatus


class ProcessingContext(BaseModel):
    """ Model for simple processing entities usage. """

    feature: FeatureModel
    scenario: ScenarioModel
    test_run: TestRunModel


class TestUserModel(sqlalchemy_to_pydantic(TestUser)):  # type: ignore
    """ Model for :class:`TestUser` row. """

    feature_type: FeatureTypeModel


class EmulationModel(sqlalchemy_to_pydantic(Emulation)):  # type: ignore
    """ Model for :class:`Emulation` row. """

    test_user: TestUserModel


class EmulationRunModel(sqlalchemy_to_pydantic(EmulationRun)):  # type: ignore
    """ Model for :class:`EmulationRun` row. """

    emulation: EmulationModel


def get_context_by_test_run_id(test_run_id: int) -> ProcessingContext:
    with create_session() as session:
        db_test_run = session.query(TestRun).filter_by(id=test_run_id).one()
        db_scenario = session.query(Scenario).filter_by(id=db_test_run.scenario.id).one()
        db_feature = session.query(Feature).filter_by(id=db_scenario.feature.id).one()
        return ProcessingContext(
            feature=FeatureModel.from_orm(db_feature),
            scenario=ScenarioModel.from_orm(db_scenario),
            test_run=TestRunModel.from_orm(db_test_run),
        )



import logging
import typing
from typing import Any

import flask
import ldap
from flask_admin import AdminIndexView, expose
from flask_login import login_required, login_user, logout_user
from sqlalchemy.exc import OperationalError, ProgrammingError
from werkzeug.wrappers import Response
from wtforms.validators import ValidationError

from overhave.admin.views.index.custom_page import CustomPageForm
from overhave.admin.views.index.login_form import LoginForm
from overhave.entities.authorization.manager import IAdminAuthorizationManager
from overhave.factory import OverhaveContext

logger = logging.getLogger(__name__)


class OverhaveIndexView(AdminIndexView):
    """ View for index. """

    def __init__(self, name: str, url: str, context: OverhaveContext, auth_manager: IAdminAuthorizationManager) -> None:
        super().__init__(
            name=name, url=url,
        )
        self._context = context
        self._auth_manager = auth_manager

    @expose('/login', methods=['GET', 'POST'])
    def login(self) -> Any:  # noqa: C901
        form = LoginForm(auth_manager=self._auth_manager)
        if not form.validate_on_submit():
            return self.render('login.html', form=form)
        try:
            user = form.get_user()
            login_user(user)
        except ValidationError:
            return form.flash_and_redirect('Incorrect username/password pair!')
        except ldap.SERVER_DOWN:
            return form.flash_and_redirect('LDAP authorization service is unreachable!')
        except OperationalError:
            logger.exception("DataBase is unreachable!")
            return form.flash_and_redirect('DataBase is unreachable!')
        except ProgrammingError:
            logger.exception("Error while trying to operate with DataBase")
            return form.flash_and_redirect('Error while trying to operate with DataBase')
        except Exception:
            logger.exception('Unknown error!')
            return form.flash_and_redirect("Unknown error!")
        next_path = flask.request.args.get('next')
        if next_path:
            logger.info("Redirect to %s...", next_path)
            return flask.redirect(next_path)
        return flask.redirect(flask.url_for('admin.index'))

    @expose('/logout', methods=['GET'])
    @login_required
    def logout(self) -> Response:
        logout_user()
        return flask.redirect(flask.url_for('admin.login'))

    @expose('/', methods=['GET', 'POST'])  # noqa: C901
    @login_required
    def index(self) -> typing.Any:  # noqa: C901
        return self.render(
            'index.html', form=CustomPageForm(template_path=self._context.admin_settings.index_template_path)
        )



import abc

from overhave.entities import Emulator, IAdminAuthorizationManager, IFeatureExtractor
from overhave.factory.context import OverhaveContext
from overhave.processing import IProcessor
from overhave.redis import RedisProducer
from overhave.scenario import FileManager
from overhave.storage import IEmulationStorage, IFeatureTypeStorage
from overhave.testing import ConfigInjector, PytestRunner


class IOverhaveFactory(abc.ABC):
    """ Factory interface for application entities resolution and usage. """

    @abc.abstractmethod
    def set_context(self, context: OverhaveContext) -> None:
        pass

    @property
    @abc.abstractmethod
    def context(self) -> OverhaveContext:
        pass

    @property
    @abc.abstractmethod
    def test_runner(self) -> PytestRunner:
        pass

    @property
    @abc.abstractmethod
    def feature_extractor(self) -> IFeatureExtractor:
        pass

    @property
    @abc.abstractmethod
    def file_manager(self) -> FileManager:
        pass

    @property
    @abc.abstractmethod
    def injector(self) -> ConfigInjector:
        pass

    @property
    @abc.abstractmethod
    def processor(self) -> IProcessor:
        pass

    @property
    @abc.abstractmethod
    def auth_manager(self) -> IAdminAuthorizationManager:
        pass

    @property
    @abc.abstractmethod
    def feature_type_storage(self) -> IFeatureTypeStorage:
        pass

    @property
    @abc.abstractmethod
    def redis_producer(self) -> RedisProducer:
        pass

    @property
    @abc.abstractmethod
    def emulation_storage(self) -> IEmulationStorage:
        pass

    @property
    @abc.abstractmethod
    def emulator(self) -> Emulator:
        pass



# flake8: noqa
from overhave.admin import OverhaveAppType, overhave_app
from overhave.base_settings import DataBaseSettings as OverhaveDBSettings
from overhave.base_settings import OverhaveLoggingSettings
from overhave.cli import group, set_config_to_context
from overhave.entities import (
    OverhaveEmulationSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveRedisSettings,
    OverhaveScenarioCompilerSettings,
    StepPrefixesModel,
    TranslitPack,
)
from overhave.entities.authorization.settings import (
    AuthorizationStrategy,
    OverhaveAdminSettings,
    OverhaveAuthorizationSettings,
    OverhaveLdapClientSettings,
)
from overhave.factory import ConsumerFactory as OverhaveConsumerFactory
from overhave.factory import OverhaveContext, OverhaveFactoryType
from overhave.factory import get_proxy_factory as overhave_factory
from overhave.redis import RedisStream as OverhaveRedisConsumerApp
from overhave.stash import OverhaveStashClientSettings, OverhaveStashProjectSettings
from overhave.testing import (
    OverhaveDescriptionManagerSettings,
    OverhaveProjectSettings,
    OverhaveTestSettings,
    get_description_manager,
)



import json
import logging
import re
from functools import cached_property
from typing import Any, Dict, List, Optional, cast

import flask
import werkzeug
from flask_admin import expose
from flask_admin.model import InlineFormAdmin
from flask_login import current_user
from markupsafe import Markup
from wtforms import Field, TextAreaField, ValidationError
from wtforms.widgets import HiddenInput

from overhave import db
from overhave.admin.views.base import ModelViewConfigured
from overhave.factory import get_proxy_factory
from overhave.storage.scenario_test_run import create_test_run

logger = logging.getLogger(__name__)


class ScenarioTextWidget(HiddenInput):
    """ Widget to override scenario view. """

    def __call__(self, field: Field, **kwargs: Any) -> Any:
        widget_name = type(self).__name__
        logger.debug("%s field id: '%s'", widget_name, field.id)
        inputs = super().__call__(field, **kwargs)
        logger.debug("%s inputs: '%s'", widget_name, inputs)
        return inputs + Markup(f'<div class="editor" ref="{field.id}" id="{field.id}-editor"></div>')


class ScenarioTextAreaField(TextAreaField):
    """ Field to override scenario view. """

    widget = ScenarioTextWidget()


class ScenarioInlineModelForm(InlineFormAdmin):
    """ Form to override scenario view. """

    form_overrides = dict(text=ScenarioTextAreaField)
    form_excluded_columns = ('created_at', 'test_runs')


class FeatureView(ModelViewConfigured):
    """ View for :class:`Feature` table. """

    can_view_details = False

    inline_models = (ScenarioInlineModelForm(db.Scenario),)
    create_template = 'feature_create.html'
    edit_template = 'feature_edit.html'

    column_list = ('id', 'name', 'feature_type', 'task', 'author', 'created_at', 'last_edited_by', 'released')
    form_excluded_columns = ('created_at', 'last_edited_by', 'released', 'versions')
    column_searchable_list = [
        'id',
        'name',
        'task',
        'last_edited_by',
    ]
    column_filters = ('name', 'feature_type', 'last_edited_by')

    _task_pattern = re.compile(r"\w+[-]\d+")

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def _validate_tasks(cls, tasks: List[str]) -> None:
        for task in tasks:
            if cls._task_pattern.match(task):
                continue
            raise ValidationError(
                f"Incorrect format of task specification: '{task}'! "
                "Supported: <PROJECT>-<NUMBER>, for example 'PRJ-1234'."
            )

    def on_model_change(self, form, model, is_created) -> None:  # type: ignore
        self._validate_tasks(model.task)
        if is_created:
            model.author = current_user.login
        model.last_edited_by = current_user.login
        model.released = False

    def on_model_delete(self, model) -> None:  # type: ignore
        if not (current_user.login == model.author or current_user.role == db.Role.admin):
            raise ValidationError('Only feature author or administrator could delete feature!')

    @cached_property
    def get_bdd_steps(self) -> Dict[str, Dict[str, List[str]]]:
        factory = get_proxy_factory()
        return {
            feature_type: factory.injector.get_steps(feature_type)
            for feature_type in factory.feature_extractor.feature_types
        }

    @property
    def browse_url(self) -> Optional[str]:
        browse_url_value = get_proxy_factory().context.project_settings.browse_url
        if browse_url_value is not None:
            return cast(str, browse_url_value.human_repr())
        return None

    @staticmethod
    def _run_test(data: Dict[str, Any], rendered: werkzeug.Response) -> werkzeug.Response:
        prefix = 'scenario-0'
        scenario_id = data.get(f'{prefix}-id')
        scenario_text = data.get(f'{prefix}-text')
        if not scenario_id or not scenario_text:
            logger.debug("Not found scenario for execution!")
            return rendered

        test_run_id = create_test_run(scenario_id=int(scenario_id), executed_by=current_user.login)
        return cast(werkzeug.Response, get_proxy_factory().processor.execute_test(test_run_id))

    @expose('/edit/', methods=('GET', 'POST'))
    def edit_view(self) -> werkzeug.Response:
        data = flask.request.form
        logger.debug("Request data:\n%s", json.dumps(data))

        rendered: werkzeug.Response = super().edit_view()
        logger.debug("EditView rendered")

        run_scenario_action = data.get('run')
        if not run_scenario_action:
            logger.debug("Show rendered EditView")
            return rendered

        logger.debug("Seen feature 'RUN' request")
        tasks = data.get('task').split(',')  # type: ignore
        self._validate_tasks(tasks=tasks)
        return self._run_test(data, rendered)



from typing import cast

import werkzeug
from flask_admin import expose
from flask_login import current_user
from wtforms import Form, ValidationError

from overhave import db
from overhave.admin.views.base import ModelViewConfigured
from overhave.admin.views.formatters import result_report_formatter


class TestRunView(ModelViewConfigured):
    """ View for :class:`TestRun` table. """

    list_template = 'test_run_list.html'
    details_template = 'test_run_detail.html'
    can_edit = True
    column_searchable_list = (
        'name',
        'executed_by',
    )
    column_list = (
        'id',
        'name',
        'start',
        'end',
        'executed_by',
        'status',
    )
    column_filters = (
        'name',
        'start',
        'executed_by',
        'status',
    )

    column_formatters = dict(status=result_report_formatter,)

    def on_model_change(self, form: Form, model: db.TestRun, is_created: bool) -> None:
        if not is_created and current_user.role != db.Role.admin:
            raise ValidationError('Only administrator could change test run data!')

    def on_model_delete(self, model: db.TestRun) -> None:
        if not (current_user.login == model.executed_by or current_user.role == db.Role.admin):
            raise ValidationError('Only test run initiator could delete test run result!')

    @expose('/details/', methods=('GET', 'POST'))
    def details_view(self) -> werkzeug.Response:
        return cast(werkzeug.Response, super().details_view())



import logging
from typing import cast

from flask import redirect
from flask_login import LoginManager
from werkzeug import Response

from overhave import db

logger = logging.getLogger(__name__)


def _load_user(user_id: int) -> db.BaseUser:
    """Return user for overlord. """
    logger.info('Get user %s', user_id)
    with db.create_session(expire_on_commit=False) as s:
        return cast(db.BaseUser, s.query(db.UserRole).filter(db.UserRole.id == user_id).first())


def _unathorized_response() -> Response:
    return redirect('/login')


def get_flask_login_manager() -> LoginManager:
    login_manager = LoginManager()
    login_manager.login_view = "login"
    login_manager.user_loader(_load_user)
    login_manager.unauthorized_handler(_unathorized_response)
    return login_manager



import json
import logging
import re
import subprocess
from functools import cached_property
from typing import Pattern

from overhave import db
from overhave.entities.converters import EmulationRunModel
from overhave.entities.settings import OverhaveEmulationSettings
from overhave.redis import EmulationTask
from overhave.storage import EmulationStorageError, IEmulationStorage

logger = logging.getLogger(__name__)


class EmulationError(Exception):
    """ Base exception for :class:`Emulator`. """


class DangerousExternalCommandError(EmulationError):
    """ Exception for dangerous command error. """


class EmulationNotEnabledError(EmulationError):
    """ Exception for not enabled emulation (```emulation_base_cmd``` not specified). """


_EMULATION_NOT_ENABLED_MSG = "Emulation not enabled: 'emulation_base_cmd' has not been specified!"


class ExternalCommandCheckMixin:
    """ Mixin for simple command-line command checking. """

    @cached_property
    def _dangerous_pattern(self) -> Pattern[str]:
        return re.compile(r"[$|><]+")

    def _check_dangerous(self, cmd: str) -> None:
        if self._dangerous_pattern.match(cmd) is not None:
            raise DangerousExternalCommandError(f"Dangerous external command: '{cmd}'")


class Emulator(ExternalCommandCheckMixin):
    """ Class for creation of emulation runs. """

    def __init__(self, storage: IEmulationStorage, settings: OverhaveEmulationSettings):
        self._storage = storage
        self._settings = settings
        if not self._settings.enabled:
            raise EmulationNotEnabledError(_EMULATION_NOT_ENABLED_MSG)

    def _initiate(self, emulation_run: EmulationRunModel) -> None:
        cmd_from_user = emulation_run.emulation.command.strip()
        self._check_dangerous(cmd_from_user)

        if not isinstance(self._settings.emulation_base_cmd, str):
            raise EmulationNotEnabledError(_EMULATION_NOT_ENABLED_MSG)

        emulation_base_cmd = self._settings.emulation_base_cmd.format(
            feature_type=emulation_run.emulation.test_user.feature_type.name
        ).split(" ")

        emulation_cmd = (
            [self._settings.emulation_core_path]
            + self._settings.emulation_prefix.format(
                address=self._settings.emulation_bind_ip,
                port=emulation_run.port,
                timeout=self._settings.wait_timeout_seconds,
            ).split(" ")
            + emulation_base_cmd
            + cmd_from_user.split(" ")
        )
        if isinstance(self._settings.emulation_postfix, str):
            emulation_cmd += self._settings.emulation_postfix.format(
                name=emulation_run.emulation.test_user.name.replace(" ", "_"),
                model=json.dumps(emulation_run.emulation.test_user.specification).replace(' ', ''),
            ).split(" ")
        logger.debug("Emulation command: %s", " ".join(emulation_cmd))
        try:
            subprocess.Popen(emulation_cmd)
        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            logger.error("Error while starting emulation process!")
            raise EmulationError(e)

    def start_emulation(self, task: EmulationTask) -> None:
        emulation_run_id = task.data.emulation_run_id
        try:
            emulation_run = self._storage.get_requested_emulation_run(emulation_run_id)
            logger.info("Try to emulate: %s", emulation_run.emulation)
            self._initiate(emulation_run)
            self._storage.set_emulation_run_status(emulation_run_id=emulation_run.id, status=db.EmulationStatus.READY)
        except (EmulationError, EmulationStorageError) as e:
            logger.exception("Could not emulate task %s!", task)
            self._storage.set_error_emulation_run(emulation_run_id=emulation_run_id, traceback=str(e))



from overhave.entities import TranslitPack

RUSSIAN_TRANSLIT_PACK = TranslitPack(
    language_code="ru",
    language_name="russian",
    pre_processor_mapping={
        u"ж": u"zh",
        u"х": u"kh",
        u"ц": u"ts",
        u"ч": u"ch",
        u"ш": u"sh",
        u"щ": u"shch",
        u"ю": u"yu",
        u"я": u"ya",
        u"ъ": u"",
        u"ь": u"",
        u" ": u"_",
        u",": u"",
        u".": u"",
        u"(": u"",
        u")": u"",
        u":": u"",
        u"?": u"",
    },
    mapping=(u"абвгдеёзийклмнопрстуфыэ₽", u"abvgdeeziyklmnoprstufye_"),
)



import logging
from typing import Dict, List, Optional

from _pytest.main import Session

from overhave.entities import FeatureTypeName, OverhaveFileSettings, StepPrefixesModel
from overhave.testing.settings import OverhaveProjectSettings
from overhave.testing.step_collector import StepCollector
from overhave.testing.test_runner import PytestRunner

logger = logging.getLogger(__name__)


class NullableEntitiesError(Exception):
    """ Application error for situation with unresolved entities. """


class ConfigInjector:
    """ Special class to operate with pytest entites, for example get data and patch objects. """

    def __init__(self) -> None:
        self._project_settings: Optional[OverhaveProjectSettings] = None
        self._step_collector: Optional[StepCollector] = None
        self._test_runner: Optional[PytestRunner] = None
        self._file_settings: Optional[OverhaveFileSettings] = None
        self._feature_types: Optional[List[FeatureTypeName]] = None

        self._current_type: Optional[FeatureTypeName] = None
        self._steps: Dict[FeatureTypeName, Dict[str, List[str]]] = {}

        self._initialized = False

    @staticmethod
    def patch_pytestbdd_prefixes(custom_step_prefixes: Optional[StepPrefixesModel]) -> None:
        """ Def for patch pytestbdd STEP_PREFIXES. """
        if custom_step_prefixes is not None:
            step_prefixes = custom_step_prefixes.extend_defaults()
            logger.debug("Successfully extended pytestbdd step prefixes with custom prefixes:\n%s", step_prefixes)

    def supplement_on_fly(
        self,
        project_settings: OverhaveProjectSettings,
        file_settings: OverhaveFileSettings,
        step_collector: StepCollector,
        test_runner: PytestRunner,
        feature_types: List[FeatureTypeName],
    ) -> None:
        """ Def for dynamical supplement of settings on-the-fly (strictly after PyTest configuring). """
        self._project_settings = project_settings
        self._step_collector = step_collector
        self._test_runner = test_runner
        self._file_settings = file_settings
        self._feature_types = feature_types

        self._steps.update({feature_type: {} for feature_type in self._feature_types})

    def collect_configs(self) -> None:
        """ Special def for pytest run in collect-only mode. Collects pytestbdd steps and other config data. """
        if self._feature_types is None or self._test_runner is None or self._file_settings is None:
            raise NullableEntitiesError
        if not self._initialized:
            logger.info("Started initialization process...")
            for feature_type in self._feature_types:
                self._current_type = feature_type
                self._test_runner.collect_only(
                    fixture_file=(
                        self._file_settings.fixtures_base_dir
                        / self._file_settings.fixtures_file_template_mask.format(feature_type=feature_type)
                    ),
                )
            self._current_type = None
            logger.info("Successfully initialized.")
            self._initialized = True

    @property
    def _inside_collecting(self) -> bool:
        return self._current_type is not None and not self._initialized

    def _load_from_session(self, session: Session) -> None:
        """ Loads pytestbdd steps and other config data from pytest Session. """
        if self._step_collector is None:
            raise NullableEntitiesError
        if self._inside_collecting:
            step_fixtures = self._step_collector.get_pytestbdd_steps(session)
            steps_dict = self._step_collector.compile_steps_dict(step_fixtures)
            if steps_dict:
                logger.debug("Loaded steps dict:\n%s", steps_dict)
            self._steps[self._current_type] = steps_dict  # type: ignore

    def adapt(self, session: Session) -> None:
        """ Main def for data loading from pytest Session and enrichment of pytest configs. """
        if self._inside_collecting:
            self._load_from_session(session)

    def get_steps(self, feature_type: FeatureTypeName) -> Dict[str, List[str]]:
        return self._steps.get(feature_type, {})



import abc
from typing import Optional

from wtforms import PasswordField, StringField

from overhave.db.users import BaseUser


class IAdminAuthorizationManager(abc.ABC):
    """ Abstract class for authorization manager. """

    @abc.abstractmethod
    def authorize_user(self, username_field: StringField, password_field: PasswordField) -> Optional[BaseUser]:
        pass



import abc
import logging
import socket
from typing import cast

import sqlalchemy.orm as so

from overhave import db
from overhave.entities import EmulationRunModel
from overhave.entities.settings import OverhaveEmulationSettings
from overhave.utils import get_current_time

logger = logging.getLogger(__name__)


class EmulationStorageError(Exception):
    """ Base Exception for :class:`EmulationStorage`. """


class NotFoundEmulationError(EmulationStorageError):
    """ Exception for situation without saved emulation.. """


class AllPortsAreBusyError(EmulationStorageError):
    """ Exception for situation when all ports are busy. """


class IEmulationStorage(abc.ABC):
    """ Abstract class for emulation runs storage. """

    @abc.abstractmethod
    def create_emulation_run(self, emulation_id: int, initiated_by: str) -> EmulationRunModel:
        pass

    @abc.abstractmethod
    def get_requested_emulation_run(self, emulation_run_id: int) -> EmulationRunModel:
        pass

    @abc.abstractmethod
    def set_emulation_run_status(self, emulation_run_id: int, status: db.EmulationStatus) -> None:
        pass

    @abc.abstractmethod
    def set_error_emulation_run(self, emulation_run_id: int, traceback: str) -> None:
        pass


class EmulationStorage(IEmulationStorage):
    """ Class for emulation runs storage. """

    def __init__(self, settings: OverhaveEmulationSettings):
        self._settings = settings

    def create_emulation_run(self, emulation_id: int, initiated_by: str) -> EmulationRunModel:
        with db.create_session() as session:
            emulation_run = db.EmulationRun(emulation_id=emulation_id, initiated_by=initiated_by)
            session.add(emulation_run)
            session.flush()
            return cast(EmulationRunModel, EmulationRunModel.from_orm(emulation_run))

    @staticmethod
    def _get_emulation_run(session: so.Session, emulation_run_id: int) -> db.EmulationRun:
        emulation_run: db.EmulationRun = session.query(db.EmulationRun).get(emulation_run_id)
        if emulation_run is not None:
            return emulation_run
        raise NotFoundEmulationError(f"Not found emulation run with ID {emulation_run_id}!")

    def _get_next_port(self, session: so.Session) -> int:
        allocated_sorted_runs = sorted(
            session.query(db.EmulationRun)
            .filter(db.EmulationRun.port.isnot(None))
            .order_by(db.EmulationRun.id.desc())
            .limit(len(self._settings.emulation_ports))
            .all(),
            key=lambda t: t.changed_at,
        )

        allocated_ports = {run.port for run in allocated_sorted_runs}
        logger.debug("Allocated ports: %s", allocated_ports)
        not_allocated_ports = set(self._settings.emulation_ports).difference(allocated_ports)
        logger.debug("Not allocated ports: %s", not_allocated_ports)
        if not_allocated_ports:
            for port in not_allocated_ports:
                if self._is_port_in_use(port):
                    continue
                return port
            logger.debug("All not allocated ports are busy!")
        for run in allocated_sorted_runs:
            if self._is_port_in_use(run.port):
                continue
            return cast(int, run.port)
        raise AllPortsAreBusyError("All ports are busy - could not find free port!")

    def _is_port_in_use(self, port: int) -> bool:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex((self._settings.emulation_bind_ip, port)) == 0

    def get_requested_emulation_run(self, emulation_run_id: int) -> EmulationRunModel:
        with db.create_session() as session:
            emulation_run = self._get_emulation_run(session, emulation_run_id)
            emulation_run.status = db.EmulationStatus.REQUESTED
            emulation_run.port = self._get_next_port(session)
            emulation_run.changed_at = get_current_time()
            return cast(EmulationRunModel, EmulationRunModel.from_orm(emulation_run))

    def set_emulation_run_status(self, emulation_run_id: int, status: db.EmulationStatus) -> None:
        with db.create_session() as session:
            emulation_run = self._get_emulation_run(session, emulation_run_id)
            if emulation_run.status != status:
                emulation_run.status = status
                emulation_run.changed_at = get_current_time()

    def set_error_emulation_run(self, emulation_run_id: int, traceback: str) -> None:
        with db.create_session() as session:
            emulation_run = self._get_emulation_run(session, emulation_run_id)
            emulation_run.status = db.EmulationStatus.ERROR
            emulation_run.traceback = traceback
            emulation_run.changed_at = get_current_time()



class PrefixMixin:
    """ Mixin for getting specified data in format of `pytest-bdd` prefix. """

    @staticmethod
    def _as_prefix(key: str) -> str:
        if ":" not in key:
            return f"{key.title()}:"
        return key



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['overhave',
 'overhave.admin',
 'overhave.admin.flask',
 'overhave.admin.views',
 'overhave.admin.views.index',
 'overhave.cli',
 'overhave.cli.db',
 'overhave.db',
 'overhave.entities',
 'overhave.entities.authorization',
 'overhave.entities.authorization.manager',
 'overhave.entities.authorization.manager.ldap',
 'overhave.entities.feature',
 'overhave.entities.language',
 'overhave.extra',
 'overhave.factory',
 'overhave.processing',
 'overhave.redis',
 'overhave.scenario',
 'overhave.stash',
 'overhave.stash.manager',
 'overhave.storage',
 'overhave.testing',
 'overhave.testing.plugin_utils',
 'overhave.testing.plugin_utils.allure_utils',
 'overhave.utils']

package_data = \
{'': ['*'], 'overhave.admin': ['files/*', 'files/ace-src/*', 'templates/*']}

install_requires = \
['Flask-WTF>=0.14.2,<0.15.0',
 'Jinja2>=2.11.2,<3.0.0',
 'WTForms>=2.2,<3.0',
 'alembic>=1.4.3,<2.0.0',
 'click>=7.0,<8.0',
 'docker',
 'flask-admin>=1.5,<2.0',
 'flask-login>=0.4.1,<0.5.0',
 'gunicorn>=20.0.4,<21.0.0',
 'httptools>=0.1.1,<0.2.0',
 'ldap3>=2.6,<3.0',
 'psycopg2-binary>=2.8,<3.0',
 'pydantic-sqlalchemy>=0.0.6,<0.0.7',
 'pydantic>=1.7',
 'python-dateutil>=2.8.1,<3.0.0',
 'python-ldap>=3.2,<4.0',
 'pytz>=2019.1,<2020.0',
 'redis>=3.4.1,<4.0.0',
 'requests>=2.0.0',
 'sqlalchemy-utc==0.10.0',
 'sqlalchemy>=1.3.3,<2.0.0',
 'sqlalchemy_utils>=0.34.0,<0.35.0',
 'tenacity',
 'walrus>=0.8.0,<0.9.0',
 'wsgi_intercept>=1.8,<2.0',
 'yarl>=1.1.1,<2.0.0']

entry_points = \
{'console_scripts': ['overhave = overhave.cli:overhave',
                     'overhave-demo = demo:overhave_demo'],
 'pytest11': ['overhave = overhave.testing.plugin']}

setup_kwargs = {
    'name': 'overhave',
    'version': '1.5.5',
    'description': 'Overhave - web-framework for BDD',
    'long_description': '========\nOverhave\n========\n\n.. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/label_img.png\n  :width: 600\n  :align: center\n  :alt: Overhave framework\n\n  `Overhave`_ is the web-framework for BDD: scalable, configurable, easy to use, based on\n  `Flask Admin`_ and `Pydantic`_.\n\n  .. image:: https://github.com/TinkoffCreditSystems/overhave/workflows/CI/badge.svg\n    :target: https://github.com/TinkoffCreditSystems/overhave/actions?query=workflow%3ACI\n    :alt: CI\n\n  .. image:: https://img.shields.io/pypi/pyversions/overhave.svg\n    :target: https://pypi.org/project/overhave\n    :alt: Python versions\n\n  .. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/TinkoffCreditSystems/overhave\n    :alt: Code style\n\n  .. image:: https://img.shields.io/pypi/v/overhave?color=%2334D058&label=pypi%20package\n    :target: https://pypi.org/project/overhave\n    :alt: Package version\n\n--------\nFeatures\n--------\n\n* Ready web-interface for easy BDD features management with `Ace`_ editor\n* Traditional Gherkin format for scenarios provided by `pytest-bdd`_\n* Execution and reporting of BDD features based on `PyTest`_  and `Allure`_\n* Auto-collection of `pytest-bdd`_ steps and display on the web-interface\n* Simple business-alike scenarios structure, easy horizontal scaling\n* Built-in wrappers for `pytest-bdd`_ hooks to supplement `Allure`_ report\n* Ability to create and use several BDD keywords dictionary with different languages\n* Versioning and deployment of scenario drafts to `Bitbucket`_\n* Built-in configurable management of users and groups permissions\n* Configurable strategy for user authorization, LDAP also provided\n* Database schema based on `SQLAlchemy`_ models and works with PostgreSQL\n* Still configurable as `Flask Admin`_, supports plug-ins and extensions\n* Distributed `producer-consumer` architecture based on Redis streams\n  through `Walrus`_ (partially integrated)\n* Web-browser emulation ability with custom toolkit (`GoTTY`_, for example)\n* Simple command-line interface, provided with `Click`_\n\n------------\nInstallation\n------------\n\nYou can install **Overhave** via pip from PyPI:\n\n.. code-block:: shell\n\n    pip install overhave\n\n--------\nOverview\n--------\n\nWeb-interface\n-------------\n\nThe web-interface is a basic tool for BDD features management. It consists of:\n\n* `Info` - index page with optional information about your tool or project;\n* `Scenarios` - section for features management, contains subsections\n    `Features`, `Test runs` and `Versions`:\n\n    * `Features`\n        gives an interface for features records management and provides info\n        about id, name author, time, editor and publishing status; it is possible\n        to search, edit or delete items through interface;\n    * `Test runs`\n        gives an interface for test runs management and provides info about\n\n        .. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/test_runs_img.png\n          :width: 500\n          :align: center\n          :alt: Script panel\n    * Versions\n        contains feature versions for corresponding to test runs.\n\n* `Access` - section for access management, contains `Users` and\n    `Groups` subsections;\n* `Emulation` - experimental section for alternative tools implementation\n    (in developing).\n\n**Overhave** features could be created and/or edited through special\n*script panel* in feature edit mode. Feature should have type registered by the\napplication, unique name, specified tasks list with the traditional format\n```PRJ-NUMBER``` and scenario text.\n\n**Script panel** has `pytest-bdd`_ steps table on the right side of interface.\nThese steps should be defined in appropriate fixture modules and registered\nat the application on start-up to be displayed.\n\n\n.. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/panel_img.png\n  :width: 600\n  :align: center\n  :alt: Script panel\n\n  Example of **Overhave** script panel in feature edit mode\n\nAllure report\n-------------\n\n**Overhave** generates `Allure`_ report after tests execution in web-interface.\nIf you execute tests manually through `PyTest`_, these results are could be\nconverted into the `Allure`_ report also with the `Allure CLI`_ tool.\nThis report contains scenarios descriptions as they are described in features.\n\n.. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/report_img.png\n  :width: 600\n  :align: center\n  :alt: Allure test-case report\n\n  Example of generated `Allure`_ report after execution of **Overhave**\'s feature\n\nCommand-line interface\n----------------------\n**Overhave** has a CLI that provides a simple way to start service web-interface,\nrun consumer and execute basic database operations. Examples are below:\n\n.. code-block:: shell\n\n    overhave db create-all\n    overhave admin --port 8080\n    overhave consumer -s EMULATION\n\n**Note**: service start-up takes a set of settings, so you can set them through\nvirtual environment with prefix ```OVERHAVE_```, for example ```OVERHAVE_DB_URL```.\nIf you want to configure settings in more explicit way through context injection,\nplease see next part of docs.\n\nContext injection\n-----------------\n\nContext setting\n^^^^^^^^^^^^^^^\n\nService could be configured via application context injection with prepared\ninstance of `OverhaveContext` object. This context could be set using\n```set_context``` function of initialized ```ProxyFactory``` instance.\n\nFor example, ```my_custom_context``` prepared. So, application start-up could\nbe realised with follow code:\n\n.. code-block:: python\n\n    from overhave import overhave_app, overhave_factory\n\n    factory = overhave_factory()\n    factory.set_context(my_custom_context)\n    overhave_app(factory).run(host=\'localhost\', port=8080, debug=True)\n\n**Note**:\n\n* ```overhave_app``` is the prepared `Flask` application with already enabled\n    Flask Admin and Login Manager plug-ins;\n* ```overhave_factory``` is a function for LRU cached instance of the **Overhave**\n    factory ```ProxyFactory```; the instance has an access to application components,\n    directly used in ```overhave_app```.\n* ```my_custom_context``` is an example of context configuration, see an\n    example code in `context_example.rst`_.\n\nImport context in PyTest\n^^^^^^^^^^^^^^^^^^^^^^^^\n\n**Overhave** has it\'s own built-in `PyTest`_ plugin, which is used to enable\nand configure injection of prepared context into application core instance.\nThe plugin provides two options:\n\n* `--enable-injection` - flag to enable context injection;\n\n* `--ctx-module` - option specifying path to Python module with context injection.\n\nThe module with context injection should contain usage of\n```set_context``` function, but this module should be\nunique and created only for `PyTest`_ usage instead of web-interface start-up.\n\nFor example, ```module_with_injection.py``` module contains:\n\n.. code-block:: python\n\n    from overhave import overhave_factory\n\n    overhave_factory().set_context(my_custom_context)\n\nAnd `PyTest` usage should be similar to:\n\n.. code-block:: bash\n\n    pytest --enable-injection --ctx-module=module_with_injection\n\nSpecified module will be imported before tests start-up (with\n```pytest_configure``` `PyTest`_ hook).\n\n\nFeatures structure\n------------------\n\n**Overhave** supports it\'s own special structure of features storage:\n\n.. image:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/features_structure_img.png\n  :width: 400\n  :alt: Features structure example\n\n**Base features directory** could contain different feature types as\nseparate directories, each of them corresponds to predefined `pytest-bdd`_\nset of steps. It is possible to create your own horizontal structure of\ndifferent product directions with unique steps and `PyTest`_ fixtures.\n\n**Note**: this structure is used in **Overhave** application. The formed data\ngives a possibility to specify registered feature type in the web-interface\n*script panel*. Also, this structure defines which steps will be displayed in\nthe rigth side of *script panel*.\n\nFeature format\n--------------\n\n**Overhave** has it\'s own special feature\'s text format, which inherits\nGherkin from `pytest-bdd`_ with small updates:\n\n* required tag that is related to existing feature type directory, where\n    current feature is located;\n* info about feature - who is creator, last editor and publisher;\n* task tracker\'s tickets with traditional format ```PRJ-NUMBER```.\n\nAn example of filled feature content is located in\n`feature_example.rst`_.\n\nLanguage\n--------\n\nThe web-interface language is ENG by default and could not be switched\n(if it\'s necessary - please, create a ```feature request``` or contribute\nyourself).\n\nThe feature text as well as `pytest-bdd`_ BDD keywords are configurable\nwith **Overhave** extra models, for example RUS keywords are already defined\nin framework and available for usage:\n\n.. code-block:: python\n\n    from overhave.extra import RUSSIAN_PREFIXES, RUSSIAN_TRANSLIT_PACK\n\n    language_settings = OverhaveLanguageSettings(\n        step_prefixes=RUSSIAN_PREFIXES,\n        translit_pack=RUSSIAN_TRANSLIT_PACK\n    )\n\n**Note**: you could create your own prefix-value mapping for your language:\n\n.. code-block:: python\n\n    from overhave import StepPrefixesModel\n\n    GERMAN_PREFIXES = StepPrefixesModel(\n        FEATURE="Merkmal:",\n        SCENARIO_OUTLINE="Szenarioübersicht:",\n        SCENARIO="Szenario:",\n        BACKGROUND="Hintergrund:",\n        EXAMPLES="Beispiele:",\n        EXAMPLES_VERTICAL="Beispiele: Vertikal",\n        GIVEN="Gegeben ",\n        WHEN="Wann ",\n        THEN="Dann ",\n        AND="Und ",\n        BUT="Aber ",\n    )\n\nCustom index\n------------\n\n**Overhave** gives an ability to set custom index.html file for rendering. Path\nto file could be set through environment as well as set with context:\n\n.. code-block:: python\n\n    admin_settings = OverhaveAdminSettings(\n        index_template_path="/path/to/index.html"\n    )\n\n\nAuthorization strategy\n----------------------\n\n**Overhave** provides several authorization strategies, declared by\n```AuthorizationStrategy``` enum:\n\n* `Simple` - strategy without real authorization.\n    Every user could use preferred name. This name will be used for user\n    authority. Every user is unique. Password not required.\n\n* `Default` - strategy with real authorization.\n    Every user could use only registered credentials.\n\n* LDAP - strategy with authorization using remote LDAP server.\n    Every user should use his LDAP credentials. LDAP\n    server returns user groups. If user in default \'admin\' group or his groups\n    list contains admin group - user will be authorized. If user already placed\n    in database - user will be authorized too. No one password stores.\n\nAppropriate strategy and additional data should be placed into\n```OverhaveAuthorizationSettings```, for example LDAP strategy could be\nconfigured like this:\n\n.. code-block:: python\n\n    auth_settings=OverhaveAuthorizationSettings(\n        auth_strategy=AuthorizationStrategy.LDAP, admin_group="admin"\n    )\n\n\n------------\nContributing\n------------\n\nContributions are very welcome.\n\nPreparation\n-----------\n\nProject installation is very easy\nand takes just few prepared commands (`make pre-init` works only for Ubuntu;\nso you can install same packages for your OS manually):\n\n.. code-block:: shell\n\n    make pre-init\n    make init\n\nPackages management is provided by `Poetry`_.\n\nCheck\n-----\n\nTests can be run with `Tox`_. `Docker-compose`_ is used for other services\npreparation and serving, such as database. Simple tests and linters execution:\n\n.. code-block:: shell\n\n    docker-compose up -d db\n    make test\n    make lint\n\nPlease, see `make` file and discover useful shortcuts. You could run tests\nin docker container also:\n\n.. code-block:: shell\n\n    make test-docker\n\nDocumentation build\n-------------------\n\nProject documentation could be built via `Sphinx`_ and simple `make` command:\n\n.. code-block:: shell\n\n    make build-docs\n\nBy default, the documentation will be built using `html` builder into `_build`\ndirectory.\n\n-------\nLicense\n-------\n\nDistributed under the terms of the `GNU GPLv2`_ license.\n\n------\nIssues\n------\n\nIf you encounter any problems, please report them here in section `Issues`\nwith a detailed description.\n\n.. _`Overhave`: https://github.com/TinkoffCreditSystems/overhave\n.. _`Pydantic`: https://github.com/samuelcolvin/pydantic\n.. _`Flask Admin`: https://github.com/flask-admin/flask-admin\n.. _`Ace`: https://github.com/ajaxorg/ace\n.. _`PyTest`: https://github.com/pytest-dev/pytest\n.. _`pytest-bdd`: https://github.com/pytest-dev/pytest-bdd\n.. _`Allure`: https://github.com/allure-framework/allure-python\n.. _`Allure CLI`: https://docs.qameta.io/allure/#_get_started\n.. _`Bitbucket`: https://www.atlassian.com/git\n.. _`SQLAlchemy`: https://github.com/sqlalchemy/sqlalchemy\n.. _`Walrus`: https://github.com/coleifer/walrus\n.. _`GoTTY`: https://github.com/yudai/gotty\n.. _`GNU GPLv2`: http://www.apache.org/licenses/LICENSE-2.0\n.. _`Tox`: https://github.com/tox-dev/tox\n.. _`Poetry`: https://github.com/python-poetry/poetry\n.. _`Docker-compose`: https://docs.docker.com/compose\n.. _`Click`: https://github.com/pallets/click\n.. _`Sphinx`: https://github.com/sphinx-doc/sphinx\n.. _`context_example.rst`: https://github.com/TinkoffCreditSystems/overhave/blob/master/docs/includes/context_example.rst\n.. _`feature_example.rst`: https://github.com/TinkoffCreditSystems/overhave/blob/master/docs/includes/features_structure_example/feature_type_1/full_feature_example_en.feature',
    'author': 'Vladislav Mukhamatnurov',
    'author_email': 'livestreamepidemz@yandex.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import abc

from overhave.stash.models import StashPrCreationResponse


class IStashProjectManager(abc.ABC):
    """ Abstract class for feature pull requests management. """

    @abc.abstractmethod
    def create_pull_request(self, test_run_id: int) -> StashPrCreationResponse:
        pass



import enum


class AuthorizationStrategy(str, enum.Enum):
    """
    Authorization strategies Enum.

    Simple - strategy without real authorization. Every user could use preferred name. This name will be used for user
    authority. Every user is unique. Password not required.
    Default - strategy with real authorization. Every user could use only registered credentials.
    LDAP - strategy with authorization using remote LDAP server. Every user should use his LDAP credentials. LDAP
    server returns user groups. If user in default 'admin' group or his groups list contains admin group - user
    will be authorized. If user already placed in database - user will be authorized too. No one password stores.
    """

    SIMPLE = "simple"
    DEFAULT = "default"
    LDAP = "ldap"



from flask_admin import Admin
from flask_admin.menu import MenuLink

from overhave.admin.views.index.view import OverhaveIndexView


def get_flask_admin(index_view: OverhaveIndexView) -> Admin:
    admin = Admin(name='Overhave', template_mode='bootstrap3', index_view=index_view,)
    admin.add_link(MenuLink(name='Log out', url='/logout'))
    return admin



# flake8: noqa
from .emulation import EmulationStorage, EmulationStorageError, IEmulationStorage
from .feature import FeatureTypeStorage, IFeatureTypeStorage
from .scenario_test_run import create_test_run, save_draft, set_report, set_run_status, set_traceback



import logging
from pathlib import Path
from typing import cast

import pytest

from overhave.testing.settings import OverhaveTestSettings

logger = logging.getLogger(__name__)


class PytestRunner:
    """ Class for running `PyTest` in test and collect-only modes. """

    def __init__(self, test_settings: OverhaveTestSettings) -> None:
        self._test_settings = test_settings

    def run(self, fixture_file: str, alluredir: str) -> int:
        pytest_cmd = [
            fixture_file,
            f"--alluredir={alluredir}",
        ]
        if isinstance(self._test_settings.default_pytest_addoptions, str):
            pytest_cmd.extend(self._test_settings.default_pytest_addoptions.split(" "))
        if isinstance(self._test_settings.extra_pytest_addoptions, str):
            pytest_cmd.extend(self._test_settings.extra_pytest_addoptions.split(" "))

        if self._test_settings.workers is not None:
            pytest_cmd.extend(["-n", f"{self._test_settings.workers}"])

        logger.debug("Prepared pytest args: %s", pytest_cmd)
        return cast(int, pytest.main(pytest_cmd))

    @staticmethod
    def collect_only(fixture_file: Path) -> None:
        logger.info("Started tests collection process with '%s'...", fixture_file.name)
        pytest_cmd = [fixture_file.as_posix(), "--collect-only", "-qq", "--disable-pytest-warnings"]
        logger.debug("Prepared pytest args: %s", pytest_cmd)
        pytest.main(pytest_cmd)
        logger.info("Finished tests collection process with '%s'.", fixture_file.name)



from datetime import datetime
from typing import Dict, Final, List, Optional, Union

from pydantic import BaseModel, Field, validator


class StashProject(BaseModel):
    """ Model for Stash pull-request project slug. """

    key: str


class StashRepository(BaseModel):
    """ Model for Stash pull-request repository. """

    name: str = Field(..., alias='slug')
    project: StashProject


class StashBranch(BaseModel):
    """ Model for Stash pull-request branch. """

    branch: str = Field(..., alias='id')
    repository: StashRepository

    @validator('branch')
    def insert_refs(cls, v: str) -> str:
        return 'refs/heads/' + v


class StashReviewerInfo(BaseModel):
    """ Model for Stash pull-request reviewer information. """

    name: str


class StashReviewer(BaseModel):
    """ Model for Stash pull-request reviewer. """

    user: StashReviewerInfo


class StashBasicPrInfo(BaseModel):
    """ Model for Stash basic pull-request information. """

    title: Optional[str]
    open: bool


class StashPrRequest(StashBasicPrInfo):
    """ Model for Stash pull-request request. """

    description: str
    state: str = "OPEN"
    closed: bool = False
    source_branch: StashBranch = Field(..., alias='fromRef')
    target_branch: StashBranch = Field(..., alias='toRef')
    locked: bool = False
    close_source_branch: bool = True
    reviewers: List[StashReviewer]


StashLinksType = Dict[str, List[Dict[str, str]]]


class StashPrCreationResponse(StashBasicPrInfo):
    """ Model for Stash pull-request creation response. """

    created_date: datetime = Field(..., alias='createdDate')
    updated_date: datetime = Field(..., alias='updatedDate')
    pull_request_url: Optional[str]
    traceback: Optional[Exception]
    links: Optional[StashLinksType]

    class Config:
        arbitrary_types_allowed = True

    def get_pr_url(self) -> str:
        if self.pull_request_url is not None:
            return self.pull_request_url
        if self.links is not None:
            return self.links['self'][0]['href']
        raise RuntimeError("Could not get pull-request URL from response!")


class StashRequestError(BaseModel):
    """ Model for Stash request error. """

    context: Optional[str]
    message: str
    exception_name: str = Field(..., alias="exceptionName")


class StashErrorResponse(BaseModel):
    """ Model for Stash error response. """

    errors: List[StashRequestError]

    @property
    def duplicate(self) -> bool:
        for error in self.errors:
            if not error.exception_name.endswith("DuplicatePullRequestException"):
                continue
            return True
        return False


AnyStashResponseModel = Union[StashPrCreationResponse, StashErrorResponse]
STASH_RESPONSE_MODELS: Final = [StashPrCreationResponse, StashErrorResponse]



class FeatureTypeExtractionError(RuntimeError):
    """ Exception for feature type extraction error. """


class ScenariosTestFileNotFound(FileNotFoundError):
    """ Exception for situation without scenarios test files. """



import click

from overhave.cli.group import overhave
from overhave.factory import ConsumerFactory, get_proxy_factory
from overhave.redis import RedisStream


@overhave.command(short_help='Run Overhave Redis consumer')
@click.option(
    "-s",
    "--stream",
    type=click.Choice(RedisStream.__members__),
    callback=lambda c, p, v: getattr(RedisStream, v),
    help="Redis stream, which defines application",
)
def consumer(stream: RedisStream) -> None:
    """ Run Overhave Redis consumer. """
    factory = get_proxy_factory()
    factory.context.logging_settings.setup_logging()
    ConsumerFactory(factory=factory, stream=stream).runner.run()



import click

from overhave.base_settings import DataBaseSettings
from overhave.cli.db.regular import create_all, drop_all, set_config_to_context
from overhave.cli.group import overhave


def _db_commands(group: click.Group) -> click.Group:
    """ Add commands to group. """
    group.add_command(create_all)
    group.add_command(drop_all)
    return group


@_db_commands
@overhave.group(short_help='Commands for simple database operations')
@click.pass_context
def db(ctx: click.Context) -> None:
    """ Commands for simple database operations. """
    set_config_to_context(context=ctx, settings=DataBaseSettings())



import logging
import re
from functools import cached_property
from typing import List, Optional

from pydantic import BaseModel
from pytest_bdd import types as default_types

from overhave.entities import IFeatureExtractor
from overhave.entities.settings import OverhaveLanguageSettings, OverhaveScenarioCompilerSettings
from overhave.scenario.errors import FeatureNameParsingError, FeatureTypeParsingError, LastEditorParsingError
from overhave.scenario.mixin import PrefixMixin

logger = logging.getLogger(__name__)
_DEFAULT_ID = 1


class FeatureInfo(BaseModel):
    """ Model for feature info keeping. """

    name: Optional[str]
    type: Optional[str]
    author: Optional[str]
    last_edited_by: Optional[str]
    tasks: Optional[List[str]]
    scenarios: Optional[str]


class ScenarioParser(PrefixMixin):
    """ Class for feature files parsing. """

    def __init__(
        self,
        compilation_settings: OverhaveScenarioCompilerSettings,
        language_settings: OverhaveLanguageSettings,
        feature_extractor: IFeatureExtractor,
        task_links_keyword: Optional[str],
    ) -> None:
        self._compilation_settings = compilation_settings
        self._language_settings = language_settings
        self._feature_extractor = feature_extractor
        self._task_links_keyword = task_links_keyword

    @cached_property
    def _feature_prefix(self) -> str:
        if self._language_settings.step_prefixes is not None:
            return self._as_prefix(self._language_settings.step_prefixes.FEATURE)
        return self._as_prefix(default_types.FEATURE)

    @cached_property
    def _task_prefix(self) -> Optional[str]:
        if isinstance(self._task_links_keyword, str):
            return self._as_prefix(self._task_links_keyword)
        return None

    def _get_name(self, name_line: str) -> str:
        name_parts = name_line.split(self._feature_prefix)
        if not name_parts:
            raise FeatureNameParsingError(f"Could not parse feature name from '{name_line}'!",)
        return name_parts[-1].strip()

    def _get_feature_type(self, tags_line: str) -> str:
        tags = tags_line.split(self._compilation_settings.tag_prefix)
        for tag in (x.strip() for x in tags if x):
            if tag not in self._feature_extractor.feature_types:
                logger.debug("Unsupported tag: %s%s", self._compilation_settings.tag_prefix, tag)
                continue
            return tag
        raise FeatureTypeParsingError(f"Could not parse feature type from '{tags_line}'!",)

    @staticmethod
    def _get_additional_info(additional_line: str, left_pointer: str, right_pointer: str) -> str:
        result = re.search(rf"({left_pointer})+\s?\w+\s?({right_pointer})+", additional_line)
        if result:
            return result.group(0).lstrip(left_pointer).rstrip(right_pointer).strip()
        raise LastEditorParsingError("Could not parse additional info from '%s'!", additional_line)

    def _get_task_info(self, task_line: str) -> List[str]:
        tasks = task_line.lstrip(self._task_prefix).split(",")
        return [x.strip() for x in tasks]

    def _parse_feature_info(self, header: str) -> FeatureInfo:
        feature_info = FeatureInfo()
        for line in header.split("\n"):
            if line.startswith(self._compilation_settings.tag_prefix):
                feature_info.type = self._get_feature_type(line)
                continue
            if line.startswith(self._feature_prefix):
                feature_info.name = self._get_name(line)
                continue
            if self._compilation_settings.created_by_prefix in line:
                feature_info.author = self._get_additional_info(
                    line,
                    left_pointer=self._compilation_settings.created_by_prefix,
                    right_pointer=self._compilation_settings.blocks_delimiter,
                )
            if self._compilation_settings.last_edited_by_prefix in line:
                feature_info.last_edited_by = self._get_additional_info(
                    line,
                    left_pointer=self._compilation_settings.last_edited_by_prefix,
                    right_pointer=self._compilation_settings.last_edited_time_delimiter,
                )
            if self._task_prefix is not None and self._task_prefix in line:
                feature_info.tasks = self._get_task_info(line)
                continue
        return feature_info

    def parse(self, feature_txt: str) -> FeatureInfo:
        blocks_delimiter = "\n\n"
        indent = "  "
        indent_substitute = "__"
        blocks = [
            block.lstrip(" \n") for block in feature_txt.replace(indent, indent_substitute).split(blocks_delimiter)
        ]
        header = blocks.pop(0)
        feature_info = self._parse_feature_info(header)
        feature_info.scenarios = blocks_delimiter.join(blocks).replace(indent_substitute, indent)
        return feature_info



from typing import Any

from flask import redirect, request, url_for
from flask_admin.contrib.sqla import ModelView
from flask_login import current_user

from overhave.admin.views.formatters import datetime_formatter, task_formatter


class ModelViewConfigured(ModelView):
    """ Base model view. """

    column_default_sort = ('id', True)
    column_display_pk = True
    can_view_details = True
    page_size = 100
    simple_list_pager = True

    column_exclude_list = ('meta',)
    column_labels = dict(
        user='Author',
        name="Name",
        feature_type="Type",
        task="Tasks",
        last_edited_by="Editor",
        released="Published",
        created_at="Creation time",
        scenario="Scenario",
        start="Start time",
        end="Finish time",
        executed_by="Initiator",
        status="Status",
        report="Report",
        traceback="Error traceback",
        feature_id="Feature ID",
        test_run_id="Test run ID",
        pr_url="Pull-request URL / Traceback",
        initiated_by="Initiator",
    )
    column_formatters = dict(
        created_at=datetime_formatter, start=datetime_formatter, end=datetime_formatter, task=task_formatter,
    )
    column_descriptions = dict(
        name="Feature header for business scenarios",
        feature_type="Type of scenarios set",
        task="Tracker tasks converted to web links when viewed",
        last_edited_by="Last editor of scenarios set",
        author="Author of scenarios set",
        released="Features repository publishing status",
        executed_by="Initiator of scenarios set test run",
        status="Test run result",
    )

    def is_accessible(self) -> Any:
        return current_user.is_authenticated

    def inaccessible_callback(self, name: str, **kwargs: Any) -> Any:
        # redirect to login page if user doesn't have access
        return redirect(url_for('admin.login', next=request.url))


class ModelViewProtected(ModelViewConfigured):
    """ Protected view. """

    can_delete = False
    can_edit = False
    can_create = False
